#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

__all__ = ("ScenarioTreeManagerSolverClientSerial",
           "ScenarioTreeManagerSolverClientPyro",
           "ScenarioTreeManagerSolverFactory")

# TODO: handle pyro as the solver manager when even when the
#       pyro scenario tree manager is used

import math
import time
import sys

from pyutilib.pyro import shutdown_pyro_components
from pyomo.opt import (SolverFactory,
                       SolverStatus,
                       TerminationCondition,
                       SolutionStatus)
from pyomo.opt.base.solvers import OptSolver
from pyomo.opt.parallel import SolverManagerFactory
from pyomo.pysp.util.config import (PySPConfigValue,
                                    PySPConfigBlock,
                                    safe_declare_common_option,
                                    safe_register_common_option)
from pyomo.pysp.util.configured_object import \
    PySPConfiguredObject
from pyomo.pysp.scenariotree.preprocessor import \
    ScenarioTreePreprocessor
from pyomo.pysp.scenariotree.server_pyro \
    import ScenarioTreeServerPyro
from pyomo.pysp.scenariotree.manager import \
    (ScenarioTreeManager,
     _ScenarioTreeManagerWorker,
     ScenarioTreeManagerClientSerial,
     ScenarioTreeManagerClientPyro,
     ScenarioTreeSolveResults)

from six import itervalues, iteritems

#
# The ScenarioTreeManagerSolver interface adds additional
# functionality to the ScenarioTreeManager manager interface
# relating to preprocessing and solving of scenarios and
# bundles.
#

class PySPFailedSolveStatus(RuntimeError):
    """This exception gets raised when one or more
    subproblem solve statuses fail basic status checks when
    processing the results of solve requests by the
    ScenarioTreeManagerSolver."""
    def __init__(self, failures, *args, **kwds):
        super(PySPFailedSolveStatus, self).__init__(*args, **kwds)
        assert type(failures) in (list, tuple)
        self.failures = tuple(failures)

class ScenarioTreeManagerSolver(PySPConfiguredObject):

    @classmethod
    def _declare_options(cls, options=None):
        if options is None:
            options = PySPConfigBlock()

        #
        # solve and I/O related
        #
        safe_declare_common_option(options,
                                   "symbolic_solver_labels")
        safe_declare_common_option(options,
                                   "solver_options")
        safe_declare_common_option(options,
                                   "solver")
        safe_declare_common_option(options,
                                   "solver_io")
        safe_declare_common_option(options,
                                   "solver_manager")
        safe_declare_common_option(options,
                                   "disable_warmstart")
        safe_declare_common_option(options,
                                   "disable_advanced_preprocessing")
        safe_declare_common_option(options,
                                   "output_solver_log")
        safe_declare_common_option(options,
                                   "keep_solver_files")
        safe_declare_common_option(options,
                                   "comparison_tolerance_for_fixed_variables")

        return options

    def __init__(self, manager, *args, **kwds):
        if self.__class__ is ScenarioTreeManagerSolver:
            raise NotImplementedError(
                "%s is an abstract class for subclassing" % self.__class__)

        super(ScenarioTreeManagerSolver, self).__init__(*args, **kwds)
        self._manager = manager

    def _solve_objects(self,
                       object_type,
                       objects,
                       ephemeral_solver_options,
                       disable_warmstart,
                       check_status,
                       async_call):

        assert object_type in ('bundles', 'scenarios')

        # queue the solves
        _async_solve_result = self._queue_object_solves(
            object_type,
            objects,
            ephemeral_solver_options,
            disable_warmstart)

        result = self.manager.AsyncResultCallback(
            lambda: self._process_solve_results(
                object_type,
                _async_solve_result.complete(),
                check_status))
        if not async_call:
            result = result.complete()
        return result

    #
    # Interface
    #

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def close(self):
        """Close the scenario tree manager solver and any
        associated objects."""
        if self.get_option("verbose"):
            print("Closing "+str(self.__class__.__name__))
        self._close_impl()

    @property
    def manager(self):
        """Return the scenario tree manager that this object
        owns"""
        return self._manager

    def solve_subproblems(self,
                          subproblems=None,
                          **kwds):
        """Solve scenarios or bundles (if they exist).

        Args:
            subproblems (list): The list of subproblem names
                to solve. The default value of :const:`None`
                indicates that all subproblems should be
                solved. Note that if the scenario tree
                contains bundles, this should be a list of
                bundles names; otherwise it should be a list
                of scenario names.
            ephemeral_solver_options (dict): A dictionary of
                solver options to override any persistent
                solver options for this set of solves only.
            disable_warmstart (bool): Disable any warmstart
                functionality available for the selected
                subproblem solvers. Default is
                :const:`False`.
            check_status (bool): Verify that all subproblem
                solves successfully completed (optimal or
                feasible solutions are loaded). Default is
                :const:`True`. This option is meant to help
                users catch errors early on and should be
                set to :const:`False` in situations where
                more advanced status handling is necessary.
            async_call (bool): When set to :const:`True`, an
                async results object is returned that allows
                the solves to be completed
                asynchronously. Default is
                :const:`False`. Note that completing an
                asynchronous solve results in modification
                of the scenario tree state on this manager
                (i.e., solutions are loaded into the
                scenarios).

        Returns:
            A :class:`ScenarioTreeSolveResults` object storing \
            basic status information for each subproblem. If \
            the :attr:`async_call` keyword is set to :const:`True` \
            then an :class:`AsyncResult` object is returned.

        Examples:
            The following lines solve all subproblems
            (automatically validating all solutions are
            optimal or feasible) and then prints a summary
            of the results.

            >>> results = sp.solve_subproblems()
            >>> results.pprint()

            The following lines do the same by first
            initiating an asynchronous solve request.

            >>> job = sp.solve_subproblems(async_call=True)
            >>> # ... do other things ... #
            >>> results = job.complete()
            >>> results.pprint()

        Raises:
            :class:`PySPFailedSolveStatus` if the \
            :attr:`check_status` keyword is :const:`True` \
            and any solves fail.
        """
        ret = None
        if self.manager.scenario_tree.contains_bundles():
            ret = self.solve_bundles(bundles=subproblems,
                                     **kwds)
        else:
            ret = self.solve_scenarios(scenarios=subproblems,
                                       **kwds)
        return ret

    def solve_scenarios(self,
                        scenarios=None,
                        ephemeral_solver_options=None,
                        disable_warmstart=False,
                        check_status=True,
                        async_call=False):
        """Solve scenarios (ignoring bundles even if they exists).

        Args:
            scenarios (list): The list of scenario names to
                solve. The default value of :const:`None`
                indicates that all scenarios should be
                solved.
            ephemeral_solver_options (dict): A dictionary of
                solver options to override any persistent
                solver options for this set of solves only.
            disable_warmstart (bool): Disable any warmstart
                functionality available for the selected
                subproblem solvers. Default is
                :const:`False`.
            check_status (bool): Verify that all subproblem
                solves successfully completed (optimal or
                feasible solutions are loaded). Default is
                :const:`True`. This option is meant to help
                users catch errors early on and should be
                set to :const:`False` in situations where
                more advanced status handling is necessary.
            async_call (bool): When set to :const:`True`, an
                async results object is returned that allows
                the solves to be completed
                asynchronously. Default is
                :const:`False`. Note that completing an
                asynchronous solve results in modification
                of the scenario tree state on this manager
                (i.e., solutions are loaded into the
                scenarios).

        Returns:
            A :class:`ScenarioTreeSolveResults` object storing \
            basic status information for each subproblem. If \
            the :attr:`async_call` keyword is set to :const:`True` \
            then an :class:`AsyncResult` object is returned.

        Examples:
            The following lines solve all scenarios
            (automatically validating all solutions are
            optimal or feasible) and then prints a summary
            of the results.

            >>> results = sp.solve_scenarios()
            >>> results.pprint()

            The following lines do the same by first
            initiating an asynchronous solve request.

            >>> job = sp.solve_scenarios(async_call=True)
            >>> # ... do other things ... #
            >>> results = job.complete()
            >>> results.pprint()

        Raises:
            :class:`PySPFailedSolveStatus` if the \
            :attr:`check_status` keyword is :const:`True` \
            and any solves fail.
        """
        return self._solve_objects('scenarios',
                                   scenarios,
                                   ephemeral_solver_options,
                                   disable_warmstart,
                                   check_status,
                                   async_call)

    def solve_bundles(self,
                      bundles=None,
                      ephemeral_solver_options=None,
                      disable_warmstart=False,
                      check_status=True,
                      async_call=False):
        """Solve bundles (they must exists).

        Args:
            bundles (list): The list of bundle names to
                solve. The default value of :const:`None`
                indicates that all bundles should be
                solved.
            ephemeral_solver_options (dict): A dictionary of
                solver options to override any persistent
                solver options for this set of solves only.
            disable_warmstart (bool): Disable any warmstart
                functionality available for the selected
                subproblem solvers. Default is
                :const:`False`.
            check_status (bool): Verify that all subproblem
                solves successfully completed (optimal or
                feasible solutions are loaded). Default is
                :const:`True`. This option is meant to help
                users catch errors early on and should be
                set to :const:`False` in situations where
                more advanced status handling is necessary.
            async_call (bool): When set to :const:`True`, an
                async results object is returned that allows
                the solves to be completed
                asynchronously. Default is
                :const:`False`. Note that completing an
                asynchronous solve results in modification
                of the scenario tree state on this manager
                (i.e., solutions are loaded into the
                scenarios).

        Returns:
            A :class:`ScenarioTreeSolveResults` object storing \
            basic status information for each subproblem. If \
            the :attr:`async_call` keyword is set to :const:`True` \
            then an :class:`AsyncResult` object is returned.

        Examples:
            The following lines solve all bundles
            (automatically validating all solutions are
            optimal or feasible) and then prints a summary
            of the results.

            >>> results = sp.solve_bundles()
            >>> results.pprint()

            The following lines do the same by first
            initiating an asynchronous solve request.

            >>> job = sp.solve_bundles(async_call=True)
            >>> # ... do other things ... #
            >>> results = job.complete()
            >>> results.pprint()

        Raises:
            :class:`PySPFailedSolveStatus` if the \
            :attr:`check_status` keyword is :const:`True` \
            and any solves fail, or :class:`RuntimeError` if
            the scenario tree was not created with bundles.
        """

        if not self.manager.scenario_tree.contains_bundles():
            raise RuntimeError(
                "Unable to solve bundles. No bundles exist")
        return self._solve_objects('bundles',
                                   bundles,
                                   ephemeral_solver_options,
                                   disable_warmstart,
                                   check_status,
                                   async_call)

    def _process_solve_results(self,
                               object_type,
                               solve_results,
                               check_status):
        """Process and load previously queued solve results."""
        assert object_type in ('bundles', 'scenarios')
        _process_function = self.manager._process_bundle_solve_result \
                            if (object_type == 'bundles') else \
                               self.manager._process_scenario_solve_result

        manager_results = ScenarioTreeSolveResults(object_type)
        failures = []
        for object_name in solve_results:

            results = solve_results[object_name]

            if self.get_option("verbose"):
                print("Processing results for %s=%s"
                      % (object_type[:-1],
                         object_name))

            start_load = time.time()
            #
            # This method is expected to:
            #  - update the dictionaries on the manager_results object
            #    (e.g., objective, gap, solution_status)
            #  - update solutions on scenario tree objects
            _process_function(
                object_name,
                results,
                manager_results=manager_results,
                allow_consistent_values_for_fixed_vars=\
                    not self.get_option("preprocess_fixed_variables"),
                comparison_tolerance_for_fixed_vars=\
                    self.get_option("comparison_tolerance_for_fixed_variables"),
                ignore_fixed_vars=\
                    self.get_option("preprocess_fixed_variables"))

            if check_status:
                if not (((manager_results.solution_status[object_name] ==
                          SolutionStatus.optimal) or \
                         (manager_results.solution_status[object_name] ==
                          SolutionStatus.feasible)) and \
                        ((manager_results.solver_status[object_name] == \
                          SolverStatus.ok) or
                         (manager_results.solver_status[object_name] == \
                          SolverStatus.warning))):
                    failures.append(object_name)
                    if self.get_option("verbose"):
                        print("Solve failed for %s=%s"
                              % (object_type[:-1], object_name))
                else:
                    if self.get_option("verbose"):
                        print("Successfully completed solve for %s=%s"
                              % (object_type[:-1], object_name))
            else:
                if self.get_option("verbose"):
                    print("Solve for %s=%s has completed. "
                          "Skipping status check."
                          % (object_type[:-1], object_name))

            if self.get_option("output_times") or \
               self.get_option("verbose"):
                print("Time loading results for %s %s=%0.2f seconds"
                      % (object_type[:-1],
                         object_name,
                         time.time() - start_load))

        if len(failures) > 0:
            print(" ** At least one of the %s failed to solve! ** "
                  % (object_type))
            print("Reporting statues for failed %s:" % (object_type))
            manager_results.pprint_status(filter_names=lambda x: x in failures)
            raise PySPFailedSolveStatus(
                failures,
                "Solve status check failed for %s %s"
                % (len(failures), object_type))

        if self.get_option("verbose"):
            manager_results.pprint(
                output_times=self.get_option("output_times"))

        if self.get_option("output_times") or \
           self.get_option("verbose"):
            if len(failures) > 0:
                print("Skipping timing statistics due to one or more "
                      "solve failures" % (object_type))
            else:
                manager_results.print_timing_summary()

        return manager_results

    #
    # Methods defined by derived class that are not
    # part of the user interface
    #

    def _close_impl(self, *args, **kwds):
        raise NotImplementedError                  #pragma:nocover

    def _queue_object_solves(self, *args, **kwds):
        raise NotImplementedError                  #pragma:nocover

#
# A partial implementation of the ScenarioTreeManagerSolver
# interface that is common to both the Serial scenario
# tree manager solver as well as the Pyro-based manager solver worker
#

class _ScenarioTreeManagerSolverWorker(ScenarioTreeManagerSolver,
                                       PySPConfiguredObject):

    @classmethod
    def _declare_options(cls, options=None):
        if options is None:
            options = PySPConfigBlock()

        # options for controlling the solver manager
        # (not the scenario tree manager)
        safe_declare_common_option(options,
                                   "solver_manager_pyro_host")
        safe_declare_common_option(options,
                                   "solver_manager_pyro_port")
        safe_declare_common_option(options,
                                   "solver_manager_pyro_shutdown")

        ScenarioTreePreprocessor._declare_options(options)

        return options

    @property
    def preprocessor(self):
        return self._preprocessor

    def __init__(self, *args, **kwds):
        if self.__class__ is _ScenarioTreeManagerSolverWorker:
            raise NotImplementedError(
                "%s is an abstract class for subclassing" % self.__class__)

        super(_ScenarioTreeManagerSolverWorker, self).__init__(*args, **kwds)
        # TODO: Does this import need to be delayed because
        #       it is in a plugins subdirectory?
        from pyomo.solvers.plugins.solvers.persistent_solver import \
            PersistentSolver

        assert self.manager is not None

        # solver related objects
        self._scenario_solvers = {}
        self._bundle_solvers = {}
        self._preprocessor = None
        self._solver_manager = None

        # there are situations in which it is valuable to snapshot /
        # store the solutions associated with the scenario
        # instances. for example, when one wants to use a warm-start
        # from a particular iteration solve, following a modification
        # and re-solve of the problem instances in a user-defined
        # callback. the following nested dictionary is intended to
        # serve that purpose. The nesting is dependent on whether
        # bundling and or phpyro is in use
        self._cached_solutions = {}
        self._cached_scenariotree_solutions = {}

        #
        # initialize the preprocessor
        #
        self._preprocessor = None
        if not self.get_option("disable_advanced_preprocessing"):
            self._preprocessor = ScenarioTreePreprocessor(self._options,
                                                          options_prefix=self._options_prefix)
        assert self._manager.preprocessor is None
        self._manager.preprocessor = self._preprocessor

        #
        # initialize the solver manager
        #
        self._solver_manager = SolverManagerFactory(
            self.get_option("solver_manager"),
            host=self.get_option('solver_manager_pyro_host'),
            port=self.get_option('solver_manager_pyro_port'))
        for scenario in self.manager.scenario_tree._scenarios:
            assert scenario._instance is not None
            solver = self._scenario_solvers[scenario.name] = \
                SolverFactory(self.get_option("solver"),
                              solver_io=self.get_option("solver_io"))
            if isinstance(solver, PersistentSolver) and \
               self.get_option("disable_advanced_preprocessing"):
                raise ValueError("Advanced preprocessing can not be disabled "
                                 "when persistent solvers are used")
            if self._preprocessor is not None:
                self._preprocessor.add_scenario(scenario,
                                                scenario._instance,
                                                solver)
        for bundle in self.manager.scenario_tree._scenario_bundles:
            solver = self._bundle_solvers[bundle.name] = \
                SolverFactory(self.get_option("solver"),
                              solver_io=self.get_option("solver_io"))
            if isinstance(solver, PersistentSolver) and \
               self.get_option("disable_advanced_preprocessing"):
                raise ValueError("Advanced preprocessing can not be disabled "
                                 "when persistent solvers are used")
            bundle_instance = \
                self.manager._bundle_binding_instance_map[bundle.name]
            if self._preprocessor is not None:
                self._preprocessor.add_bundle(bundle,
                                              bundle_instance,
                                              solver)

    #
    # Override some methods for ScenarioTreeManager that
    # were implemented by _ScenarioTreeManagerWorker:
    #

    def _close_impl(self):
        if (self._manager is not None) and \
           (self._manager.preprocessor is not None):
            assert self.preprocessor is self._manager.preprocessor
            for bundle in self.manager.scenario_tree._scenario_bundles:
                self._preprocessor.remove_bundle(bundle)
            for scenario in self.manager.scenario_tree._scenarios:
                assert scenario._instance is not None
                self._preprocessor.remove_scenario(scenario)
            self._manager.preprocessor = None
            self._preprocessor = None
        else:
            assert self._preprocessor is None

        if self._solver_manager is not None:
            #self._solver_manager.deactivate()
            self._solver_manager = None
            if self.get_option("solver_manager_pyro_shutdown"):
                print("Shutting down Pyro components for solver manager.")
                shutdown_pyro_components(
                    host=self.get_option("solver_manager_pyro_host"),
                    port=self.get_option("solver_manager_pyro_port"),
                    num_retries=0,
                    caller_name=self.__class__.__name__)
        #for solver in self._scenario_solvers.values():
        #    solver.deactivate()
        self._scenario_solvers = {}
        #for solver in self._bundle_solvers.values():
        #    solver.deactivate()
        self._bundle_solvers = {}
        self._preprocessor = None
        self._objective_sense = None

    #
    # Abstract methods for ScenarioTreeManagerSolver:
    #

    def _queue_object_solves(self,
                             object_type,
                             objects,
                             ephemeral_solver_options,
                             disable_warmstart):

        if self.get_option("verbose"):
            print("Queuing %s solves" % (object_type[:-1]))

        assert object_type in ('bundles', 'scenarios')

        solver_dict = None
        instance_dict = None
        modify_kwds_func = None
        if object_type == 'bundles':
            if objects is None:
                objects = self.manager.scenario_tree._scenario_bundle_map
            solver_dict = self._bundle_solvers
            instance_dict = self.manager._bundle_binding_instance_map
            for bundle_name in objects:
                for scenario_name in self.manager.scenario_tree.\
                    get_bundle(bundle_name).\
                       scenario_names:
                    self.manager.scenario_tree.get_scenario(scenario_name).\
                        _instance_objective.deactivate()
            if self.preprocessor is not None:
                self.preprocessor.preprocess_bundles(bundles=objects)
                modify_kwds_func = self.preprocessor.modify_bundle_solver_keywords
        else:
            if objects is None:
                objects = self.manager.scenario_tree._scenario_map
            solver_dict = self._scenario_solvers
            instance_dict = self.manager._instances
            if self.manager.scenario_tree.contains_bundles():
                for scenario_name in objects:
                    self.manager.scenario_tree.get_scenario(scenario_name).\
                        _instance_objective.activate()
            if self.preprocessor is not None:
                self.preprocessor.preprocess_scenarios(scenarios=objects)
                modify_kwds_func = self.preprocessor.modify_scenario_solver_keywords
        assert solver_dict is not None
        assert instance_dict is not None

        # setup common solve keywords
        common_kwds = {}
        common_kwds['tee'] = self.get_option("output_solver_log")
        common_kwds['keepfiles'] = self.get_option("keep_solver_files")
        common_kwds['symbolic_solver_labels'] = \
            self.get_option("symbolic_solver_labels")
        # we always manually load solutions, so we can
        # control error reporting and such
        common_kwds['load_solutions'] = False

        # Load solver options
        solver_options = {}
        if type(self.get_option("solver_options")) is tuple:
            solver_options.update(
                OptSolver._options_string_to_dict(
                    "".join(self.get_option("solver_options"))))
        else:
            solver_options.update(self.get_option("solver_options"))
        common_kwds['options'] = solver_options

        #
        # override "persistent" values that are included from this
        # classes registered options
        #
        if ephemeral_solver_options is not None:
            common_kwds['options'].update(ephemeral_solver_options)

        # maps action handles to subproblem names
        action_handle_data = {}
        for object_name in objects:

            if modify_kwds_func is not None:
                # be sure to modify a copy of the kwds
                solve_kwds = modify_kwds_func(object_name, dict(common_kwds))
            else:
                solve_kwds = common_kwds
            opt = solver_dict[object_name]
            instance = instance_dict[object_name]
            if (not self.get_option("disable_warmstart")) and \
               (not disable_warmstart) and \
               opt.warm_start_capable():
                new_action_handle = \
                    self._solver_manager.queue(instance,
                                               opt=opt,
                                               warmstart=True,
                                               **solve_kwds)
            else:
                new_action_handle = \
                    self._solver_manager.queue(instance,
                                               opt=opt,
                                               **solve_kwds)

            action_handle_data[new_action_handle] = object_name

        return self.manager.AsyncResult(
            self._solver_manager,
            action_handle_data=action_handle_data)

class ScenarioTreeManagerSolverClientSerial(
        _ScenarioTreeManagerSolverWorker,
        ScenarioTreeManagerSolver,
        PySPConfiguredObject):

    @classmethod
    def _declare_options(cls, options=None):
        if options is None:
            options = PySPConfigBlock()
        return options

    def __init__(self, *args, **kwds):
        super(ScenarioTreeManagerSolverClientSerial, self).\
            __init__(*args, **kwds)

    #
    # Override some methods for ScenarioTreeManager that
    # were implemented by _ScenarioTreeManagerWorkerSolver:
    #

    def _close_impl(self):
        super(ScenarioTreeManagerSolverClientSerial, self)._close_impl()

    #
    # Abstract methods for ScenarioTreeManagerSolver:
    #

    # implemented by _ScenarioTreeManagerSolverWorker
    #def _queue_object_solves(...)

class ScenarioTreeManagerSolverClientPyro(ScenarioTreeManagerSolver,
                                          PySPConfiguredObject):

    @classmethod
    def _declare_options(cls, options=None):
        if options is None:
            options = PySPConfigBlock()
        return options

    default_registered_worker_name = 'ScenarioTreeManagerSolverWorkerPyro'

    #
    # Override the PySPConfiguredObject register_options implementation so
    # that the default behavior will be to register this classes default
    # worker type options along with the options for this class
    #

    @classmethod
    def register_options(cls, *args, **kwds):
        registered_worker_name = \
            kwds.pop('registered_worker_name',
                     cls.default_registered_worker_name)
        options = super(ScenarioTreeManagerSolverClientPyro, cls).\
                  register_options(*args, **kwds)
        if registered_worker_name is not None:
            worker_type = ScenarioTreeServerPyro.\
                          get_registered_worker_type(registered_worker_name)
            worker_type.register_options(options, **kwds)

        return options

    def __init__(self, manager, *args, **kwds):
        worker_registered_name = \
            kwds.pop('registered_worker_name',
                     self.default_registered_worker_name)
        super(ScenarioTreeManagerSolverClientPyro, self).\
            __init__(manager, *args, **kwds)

        assert self.manager is not None
        assert self.manager._action_manager is not None
        self._pyro_worker_map = {}
        self._pyro_base_worker_map = {}

        worker_class = ScenarioTreeServerPyro.\
                       get_registered_worker_type(
                           self.default_registered_worker_name)
        try:
            worker_options = worker_class.\
                             extract_user_options_to_dict(
                                 self._options,
                                 source_options_prefix=self._options_prefix,
                                 sparse=True)
        except KeyError:
            raise KeyError(
                "Unable to serialize options for registered worker "
                "name %s (class=%s). The worker options did not "
                "seem to match the registered options on the worker "
                "class. Did you forget to register them? Message: %s"
                % (worker_registered_name,
                   worker_class.__name__,
                   str(sys.exc_info()[1])))

        assert not self.manager._transmission_paused
        if not self.manager.get_option("pyro_handshake_at_startup"):
            self.manager.pause_transmit()
        action_handle_data = {}
        for base_worker_name in self.manager.worker_names:
            server_name = \
                self.manager.get_server_for_worker(base_worker_name)

            worker_name = base_worker_name+"_solver"
            if self.manager.get_option("verbose"):
                print("Initializing worker with name %s on "
                      "scenario tree server %s"
                      % (worker_name, server_name))

            action_handle = self.manager._action_manager.queue(
                queue_name=server_name,
                action="ScenarioTreeServerPyro_initialize",
                worker_type=worker_registered_name,
                worker_name=worker_name,
                init_args=(base_worker_name,),
                init_kwds=worker_options,
                generate_response=True)

            if self.manager.get_option("pyro_handshake_at_startup"):
                action_handle_data[worker_name] =  \
                    self.manager.AsyncResult(
                        self.manager._action_manager,
                        action_handle_data=action_handle).complete()
            else:
                action_handle_data[action_handle] = worker_name

            self._pyro_worker_map[base_worker_name] = worker_name
            self._pyro_base_worker_map[worker_name] = base_worker_name

        if not self.manager.get_option("pyro_handshake_at_startup"):
            self.manager.unpause_transmit()

        if self.manager.get_option("pyro_handshake_at_startup"):
            result = self.manager.AsyncResult(
                None,
                result=action_handle_data)
        else:
            result = self.manager.AsyncResult(
                self.manager._action_manager,
                action_handle_data=action_handle_data)
        result.complete()

    #
    # Abstract methods for ScenarioTreeManagerSolver:
    #

    def _close_impl(self):
        # release the workers created by this manager solver
        if len(self._pyro_worker_map):
            assert len(self._pyro_worker_map) == \
                len(self._pyro_base_worker_map)
            if self.manager._transmission_paused:
                print("Unpausing pyro transmissions in "
                      "preparation for releasing solver workers")
                self.manager.unpause_transmit()
            self.manager.pause_transmit()
            action_handles = []
            for base_worker_name in self._pyro_worker_map:
                worker_name = self._pyro_worker_map[base_worker_name]
                server_name = self.manager.\
                              get_server_for_worker(base_worker_name)
                action_handles.append(
                    self.manager._action_manager.queue(
                    queue_name=server_name,
                    action="ScenarioTreeServerPyro_release",
                    worker_name=worker_name,
                    generate_response=True))
            self.manager.unpause_transmit()
            self.manager._action_manager.wait_all(action_handles)
            for ah in action_handles:
                self.manager._action_manager.get_results(ah)
        self._pyro_worker_map = {}
        self._pyro_base_worker_map = {}

    def _queue_object_solves(self,
                             object_type,
                             objects,
                             ephemeral_solver_options,
                             disable_warmstart):

        assert object_type in ('bundles', 'scenarios')

        if self.get_option("verbose"):
            print("Transmitting solve requests for %s" % (object_type))

        worker_names = None
        worker_map = {}
        if objects is not None:
            if object_type == 'bundles':
                _get_worker_func = self.manager.get_worker_for_bundle
            else:
                assert object_type == 'scenarios'
                _get_worker_func = self.manager.get_worker_for_scenario
            for object_name in objects:
                worker_name = _get_worker_func(object_name)
                if worker_name not in worker_map:
                    worker_map[worker_name] = []
                worker_map[worker_name].append(object_name)
            worker_names = worker_map
        else:
            worker_names = self.manager._pyro_worker_list

        was_paused = self.manager.pause_transmit()
        action_handle_data = {}
        for base_worker_name in worker_names:
            method_args=(object_type,
                         worker_map.get(base_worker_name, None),
                         ephemeral_solver_options,
                         disable_warmstart)
            worker_name = self._pyro_worker_map[base_worker_name]
            server_name = \
                self.manager.get_server_for_worker(base_worker_name)
            action_handle_data[self.manager._action_manager.queue(
                    queue_name=server_name,
                    worker_name=worker_name,
                    action="_solve_objects_for_client",
                    args=method_args,
                    kwds={},
                    generate_response=True)] = worker_name
        if not was_paused:
            self.manager.unpause_transmit()

        return self.manager.AsyncResult(
            self.manager._action_manager,
            action_handle_data=action_handle_data,
            map_result=(lambda ah_to_result: \
                        dict((key, result[key])
                             for result in itervalues(ah_to_result)
                             for key in result)))

def ScenarioTreeManagerSolverFactory(sp, *args, **kwds):
    """Return a scenario tree manager solver appropriate for
    the provided argument.

    Args:
        sp: a serial or pyro client scenario tree manager
        *args: A single additional argument can be provided
            that is a block of registered options used to
            initialize the returned manager solver. The
            block of options can be created by calling
            :attr:`ScenarioTreeManagerSolverFactory.register_options`.
        **kwds: Additional keywords are passed to the
            manager solver that is created.

    Returns: A :class:`ScenarioTreeManagerSolver` object.

    Example:
        The preferred way to use a ScenarioTreeManagerSolver
        object is through a :const:`with` block as it
        modifies the state of the underlying scenario tree
        manager. If used outside a :const:`with` block, the
        manager solver should be shutdown by calling the
        :attr:`close` method.

        >>> with ScenarioTreeManagerSolverFactory(sp) as manager:
        >>>    results = manager.solve_subproblems()
        >>> results.pprint()

        Note that asynchronous solves should be completed
        before the manager solver is closed; otherwise the
        results are undefined.

        >>> with ScenarioTreeManagerSolverFactory(sp) as manager:
        >>>    job = manager.solve_subproblems(async_call=True)
        >>>    reuslts = job.complete()
        >>> results.pprint()
    """
    if isinstance(sp, ScenarioTreeManagerClientSerial):
        manager_class = ScenarioTreeManagerSolverClientSerial
    elif isinstance(sp, ScenarioTreeManagerClientPyro):
        manager_class = ScenarioTreeManagerSolverClientPyro
    else:
        raise ValueError("Unrecognized type for first argument: %s "
                         % (type(sp)))
    if len(args) == 0:
        options = manager_class.register_options()
    elif len(args) == 1:
        options = args[0]
    else:
        raise ValueError("At most 2 arguments allowed "
                         "for function call")
    return manager_class(sp, options, **kwds)

def _register_scenario_tree_manager_solver_options(*args, **kwds):
    if len(args) == 0:
        options = PySPConfigBlock()
    else:
        if len(args) != 1:
            raise TypeError(
                "register_options(...) takes at most 1 argument (%s given)"
                % (len(args)))
        options = args[0]
        if not isinstance(options, PySPConfigBlock):
            raise TypeError(
                "register_options(...) argument must be of type PySPConfigBlock, "
                "not %s" % (type(options).__name__))

    ScenarioTreeManagerSolverClientSerial.register_options(options,
                                                           **kwds)
    ScenarioTreeManagerSolverClientPyro.register_options(options,
                                                         **kwds)

    return options

ScenarioTreeManagerSolverFactory.register_options = \
    _register_scenario_tree_manager_solver_options



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

# May 2015 issue: reals really should not be *fixed* to compute an inner bound
# It seems like a lot to change their bounds, but that is probably the way to go.

from __future__ import division

import os
import logging
import copy

import pyomo.common.plugin
from pyomo.pysp import phextension
from pyomo.core.base import minimize
from pyomo.opt import UndefinedData

from operator import itemgetter
from six import iteritems

from pyomo.pysp.phboundbase import (_PHBoundBase,
                                    ExtractInternalNodeSolutionsforInner)

class _PHBoundExtensionImpl(_PHBoundBase):

    def __init__(self):

        _PHBoundBase.__init__(self)

    def _iteration_k_bound_solves(self, ph, storage_key):

        # storage key is for results (e.g. the ph iter number)

        # ** Some code might use the values stored in the scenario solutions
        #    to perform a weighted vote in the case of discrete
        #    variables, so it is important that we execute this
        #    before perform any new subproblem solves.

        # candidate_sol is sometimes called xhat
        try:
            candidate_sol = ExtractInternalNodeSolutionsforInner(ph)
        except:
            print("Failed to extract candiate xhat for "
                  "inner bound computation using xhat_method %s. "
                  "Skipping inner bound computation."
                  % (ph._xhat_method))
            candidate_sol = None

        # Caching the current set of ph solutions so we can restore
        # the original results. We modify the scenarios and re-solve -
        # which messes up the warm-start, which can seriously impact
        # the performance of PH. plus, we don't want lower bounding to
        # impact the primal PH in any way - it should be free of any
        # side effects.
        self.CachePHSolution(ph)

        # Save the current fixed state and fix queue.
        self.RelaxPHFixedVariables(ph)

        # Assuming the weight terms are already active but proximal
        # terms need to be deactivated deactivate all proximal terms
        # and activate all weight terms.
        self.DeactivatePHObjectiveProximalTerms(ph)

        if candidate_sol is not None:
            # Deactivate the weight terms.
            self.DeactivatePHObjectiveWeightTerms(ph)

            # Fix all non-leaf stage variables involved
            # in non-anticipativity conditions to the most
            # recently computed xbar (or something like it)
            # integers should be truly fixed, but reals require special care
            self.FixScenarioTreeVariables(ph, candidate_sol)

            # now change over to finding a feasible incumbent.
            if ph._verbose:
                print("Computing objective %s bound" %
                      ("inner" if self._is_minimizing else "outer"))

            failures = ph.solve_subproblems(warmstart=not ph._disable_warmstarts,
                                            exception_on_failure=False)

            if len(failures):

                print("Failed to compute %s bound at xhat due to "
                      "one or more solve failures" %
                  ("inner" if self._is_minimizing else "outer"))
                self._inner_bound_history[storage_key] = \
                    float('inf') if self._is_minimizing else float('-inf')
                self._inner_status_history[storage_key] = self.STATUS_SOLVE_FAILED

            else:

                if ph._verbose:
                    print("Successfully completed PH bound extension "
                          "fixed-to-xhat solves for iteration %s\n"
                          "- solution statistics:\n" % (storage_key))
                    if ph._scenario_tree.contains_bundles():
                        ph.report_bundle_objectives()
                    ph.report_scenario_objectives()

                # Compute the inner bound on the objective function.
                IBval, IBstatus = self.ComputeInnerBound(ph, storage_key)
                self._inner_bound_history[storage_key] = IBval
                self._inner_status_history[storage_key] = IBstatus

            # Undo FixScenarioTreeVariables
            self.RestoreLastPHChange(ph)

            # Undo DeactivatePHObjectiveWeightTerms
            self.RestoreLastPHChange(ph)

        else:

            if self._is_minimizing:
                self._inner_bound_history[storage_key] = float('inf')
            else:
                self._inner_bound_history[storage_key] = float('-inf')
            self._inner_status_history[storage_key] = self.STATUS_NONE

        # push the updated inner bound to PH, for reporting purposes.
        ph._update_reported_bounds(inner = self._inner_bound_history[storage_key])

        # It is possible weights have not been pushed to instance
        # parameters (or transmitted to the phsolverservers) at this
        # point.
        ph._push_w_to_instances()

        failures = ph.solve_subproblems(warmstart=not ph._disable_warmstarts,
                                        exception_on_failure=False)

        if len(failures):

            print("Failed to compute duality-based bound due to "
                  "one or more solve failures")
            self._outer_bound_history[storage_key] = \
                float('-inf') if self._is_minimizing else float('inf')
            self._outer_status_history[storage_key] = self.STATUS_SOLVE_FAILED

        else:

            if ph._verbose:
                print("Successfully completed PH bound extension "
                      "weight-term only solves for iteration %s\n"
                      "- solution statistics:\n" % (storage_key))
                if ph._scenario_tree.contains_bundles():
                    ph.report_bundle_objectives()
                ph.report_scenario_objectives()

            # Compute the outer bound on the objective function.
            self._outer_bound_history[storage_key], \
                self._outer_status_history[storage_key] = \
                    self.ComputeOuterBound(ph, storage_key)

            ph._update_reported_bounds(outer = self._outer_bound_history[storage_key])

        # Restore ph to its state prior to entering this method (e.g.,
        # fixed variables, scenario solutions, proximal terms)
        self.RestorePH(ph)

    ############ Begin Callback Functions ##############

    def reset(self, ph):
        """Invoked to reset the state of a plugin to that of post-construction"""
        self.__init__()

    def pre_ph_initialization(self,ph):
        """
        Called before PH initialization.
        """
        pass

    def post_instance_creation(self, ph):
        """
        Called after PH initialization has created the scenario
        instances, but before any PH-related
        weights/variables/parameters/etc are defined!
        """
        pass

    def post_ph_initialization(self, ph):
        """
        Called after PH initialization
        """

        if ph._verbose:
            print("Invoking post initialization callback "
                  "in phboundextension")

        self._is_minimizing = True if (ph._objective_sense == minimize) else False
        # TODO: Check for ph options that may not be compatible with
        #       this plugin and warn / raise exception

        # grab the update interval from the environment variable, if
        # it exists.
        update_interval_variable_name = "PHBOUNDINTERVAL"
        update_interval_file_name = "PHB_.DAT"
        if os.path.isfile(update_interval_file_name):
            print("phboundextension is getting the update interval from file=",
                   update_interval_file_name)
            with open(update_interval_file_name) as ifile:
                ifileval = ifile.read()
            if isinstance(ifileval, int):
                print ("update interval=",ifileval)
                self._update_interval = ifileval
            else:
                raise RuntimeError("The value must be of type integer, but the value read="+str(ifileval))

        elif update_interval_variable_name in os.environ:
            self._update_interval = int(os.environ[update_interval_variable_name])
            print("phboundextension using update interval="
                  +str(self._update_interval)+", extracted from "
                  "environment variable="+update_interval_variable_name)
        else:
            print("phboundextension using default update "
                  "interval="+str(self._update_interval))

    def post_iteration_0_solves(self, ph):
        """
        Called after the iteration 0 solves
        """

        if ph._verbose:
            print("Invoking post iteration 0 solve callback "
                  "in phboundextension")

        if ph._ph_warmstarted:
            print("PH warmstart detected. Bound computation requires solves "
                  "after iteration 0.")
            self.pre_iteration_k_solves(ph)
            return

        # Always compute a lower/upper bound here because it requires
        # no work.  The instances (or bundles) have already been
        # solved with the original (non-PH-augmented) objective and
        # are loaded with results.

        #
        # Note: We will still obtain a bound using the weights
        #       computed from PH iteration 0 in the
        #       pre_iteration_k_solves callback.
        #
        ph_iter = None

        # Note: It is important that the mipgap is not adjusted
        #       between the time after the subproblem solves
        #       and before now.
        self._outer_bound_history[ph_iter], \
            self._outer_status_history[ph_iter] = \
               self.ComputeOuterBound(ph, ph_iter)

        # dlw May 2016: the reported bound gets set for general iterations right after
        #               assignment to the history, so we do it here also
        ph._update_reported_bounds(outer = self._outer_bound_history[ph_iter])

    def post_iteration_0(self, ph):
        """
        Called after the iteration 0 solves, averages computation, and weight computation
        """
        pass

    def pre_iteration_k_solves(self, ph):
        """
        Called immediately before the iteration k solves
        """

        if ph._verbose:
            print("Invoking pre iteration k solve callback "
                  "in phboundextension")

        #
        # Note: We invoke this callback pre iteration k in order to
        #       obtain a PH bound using weights computed from the
        #       PREVIOUS iteration's scenario solutions (including
        #       those of iteration zero).
        #
        ph_iter = ph._current_iteration-1

        if (ph_iter % self._update_interval) != 0:
            return

        self._iteration_k_bound_solves(ph, ph_iter)

    def post_iteration_k_solves(self, ph):
        """
        Called after the iteration k solves!
        """
        pass

    def post_iteration_k(self, ph):
        """
        Called after the iteration k is finished, after weights have been updated!
        """
        pass

    def post_ph_execution(self, ph):
        """
        Called after PH has terminated!
        """

        if ph._verbose:
            print("Invoking post execution callback in phboundextension")

        #
        # Note: We invoke this callback in order to compute a bound
        #       using the weights obtained from the final PH
        #       iteration.
        #
        ph_iter = ph._current_iteration
        self._iteration_k_bound_solves(ph, ph_iter)

        self.ReportBoundHistory()
        self.ReportBestBound()

class phboundextension(pyomo.common.plugin.SingletonPlugin, _PHBoundExtensionImpl):

    pyomo.common.plugin.implements(phextension.IPHExtension)

    pyomo.common.plugin.alias("phboundextension")

    def __init__(self):

        _PHBoundExtensionImpl.__init__(self)



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

# dlw Dec 2014: all 'wb' now 'wt'
# dlw Jan 2015: even more 'w' and 'wb' now 'wt'
#
# ph extension for dual decomposition
#

import csv
import sys
import itertools
from operator import itemgetter
import os
thisfile = os.path.abspath(__file__)

import pyomo.common.plugin
from pyomo.core import *
from pyomo.core.base.var import _VarData
from pyomo.core.base.piecewise import _PiecewiseData
from pyomo.pysp import phextension
from pyomo.pysp.phsolverserver import _PHSolverServer
from pyomo.pysp.phsolverserverutils import \
    transmit_external_function_invocation_to_worker

from six import iteritems

# This must be in your path
_ddsip_help = 'DDSIPhelp_x64'
_ddsip_help_output = "rows+cols"
_ddsip_exe = 'DDSIP_126_x64'
_precision_string = '.17g'

_disable_stoch_rhs_flagname = "DDSIP_DISABLE_STOCHASTIC_RHS"
_disable_stoch_rhs_default = False

_disable_stoch_matrix_flagname = "DDSIP_DISABLE_STOCHASTIC_MATRIX"
_disable_stoch_matrix_default = True

_disable_stoch_costs_flagname = "DDSIP_DISABLE_STOCHASTIC_COSTS"
_disable_stoch_costs_default = True

# TODO:
# handle the case where a variable on the scenario tree does not
#  appear in the LP file
# test with range constraints
# test with == constraints
# test with <= constraints
# test with >= constraints
# test with Piecewise
# test that SOSConstraint throws error
# test number of stochastic entries match
#  (this is not a necessary or sufficient condition
#   for checking that something has gone wrong, but
#   it should catch most issues)

# POTENTIAL ISSUES
# The fact that the constraint matrix is sparse presents a problem when
# stochastic matrix coefficients are sometimes zero. In these cases, somewhere
# in the pyomo toolchain these entries will be simplified out of the constraint
# matrix (e.g, when generating the pyomo expression or when writing the
# LP file lines). In such a case, the generated stochastic matrix file might be
# garbage (but not cause an error).
#
# In a related case, if a constraint becomes "trivial" (due to stochastic coefficients)
# that could cause it appear in some scenario models but not others. Again, this
# could possibly result in garbage DDSIP inputs that don't cause an error.
#
# My proposed solution is choose a global reference scenario that all other scenarios
# must compare nonzero locations against as they write their own file. This is not
# implemented yet.

def Write_DDSIP_Input(ph, scenario_tree, scenario_or_bundle,
                      scenario_name,
                      scenario_index,
                      firststage_var_suffix,
                      global_reference_scenario=False):
    dd = DDSIP_Input(ph, scenario_tree, scenario_or_bundle,
                     scenario_name,
                     scenario_index,
                     firststage_var_suffix)
    return dd.write(ph, global_reference_scenario=global_reference_scenario)

class DDSIP_Input(object):

    def __init__(self,
                 ph, scenario_tree, scenario_or_bundle,
                 scenario_name,
                 scenario_index,
                 firststage_var_suffix):

        self.input_file_name_list = []
        self._scenario_name = scenario_name
        self._scenario_index = scenario_index
        self._lpfilename = "ddlpfile.lp."+self._scenario_name
        self._rhsfilename = "rhs.sc."+self._scenario_name
        self._matfilename = "matrix.sc."+self._scenario_name
        self._objfilename = "cost.sc."+self._scenario_name
        self._firststage_var_suffix = firststage_var_suffix
        self._reference_scenario = ph._scenario_tree.get_scenario(scenario_name)
        self._reference_scenario_instance = None

        self._FirstStageVars = []
        self._FirstStageVarIdMap = {}
        self._FirstStageDerivedVars = []
        self._FirstStageDerivedVarIdMap = {}
        self._SecondStageVars = []
        self._SecondStageVarIdMap = {}
        self._StageCostVarStageMap = {}

        # If these values remain as None, it will
        # indicate the user has explicitely set
        # disable flags on the model (which
        # we will report for confirmation purposes)
        self._num_stochastic_rhs = None
        self._num_stochastic_matrix_entries = None
        self._num_stochastic_costs = None

        self._num_first_stage_constraints = None
        self._num_second_stage_constraints = None

        # If this constraint appears in the LP file it means the
        # objective included constant terms
        self._count_constraint_ONE_VAR_CONSTANT = 0

        self._AllVars = []

        # Read from siphelp output. Includes cplex row, column
        # assignments Required for stochastic matrix file
        self._ColumnMap = {}
        self._ColumnNameList = []
        self._RowMap = {}
        self._RowNameList = []

        self._reference_scenario_instance = self._reference_scenario._instance

    def write(self, ph, global_reference_scenario=False):

        scenario_instance = self._reference_scenario_instance

        # This is usuall called just prior to solving the instances,
        # however the scenariotree uses StandardRepn for determining
        # constraint stage
        ph._preprocess_scenario_instances()

        # Write the lp file for an arbitrary scenario, also obtain the
        # maximum row/column symbol length (in characters)
        max_name_len, symbol_map = self._write_reference_scenario_lp(ph)

        # variables will stored by their name in the LP file,
        # which may be different from what is stored on the
        # ph scenario tree
        StageToConstraintMap = self._Constraints_Stages(ph, symbol_map)
        self._Populate_StageVars(ph, symbol_map)

        if global_reference_scenario:
            assert os.path.exists(self._lpfilename)
            try:
                os.remove(os.path.join(os.getcwd(),_ddsip_help_output))
            except OSError:
                pass
            try:
                os.remove(os.path.join(os.getcwd(),_ddsip_help_output+".gz"))
            except OSError:
                pass
            assert not os.path.exists(os.path.join(os.getcwd(),_ddsip_help_output))
            assert not os.path.exists(os.path.join(os.getcwd(),_ddsip_help_output+".gz"))
            print(("COMMAND= "+str(_ddsip_help)+' '+self._lpfilename+' '+str(max_name_len)))
            os.system(_ddsip_help+' '+self._lpfilename+' '+str(max_name_len))
            assert os.path.exists(os.path.join(os.getcwd(),_ddsip_help_output+".gz"))
            os.system("gzip -df "+os.path.join(os.getcwd(),_ddsip_help_output+".gz"))
        else:
            assert os.path.exists(os.path.join(os.getcwd(),_ddsip_help_output))

        dashcount = 0
        varcount = 0
        concount = 0
        with open(_ddsip_help_output,'r') as f:
            for line in f.readlines():
                pieces = line.split()
                if dashcount == 3:
                    break
                elif pieces[0] =='__________________________':
                    dashcount = dashcount+1
                elif (dashcount == 1) and (len(pieces) == 2):
                    varcount += 1
                    self._ColumnMap[pieces[1]] = int(pieces[0])
                    self._ColumnNameList.append(pieces[1])
                    assert len(self._ColumnNameList)-1 == \
                        self._ColumnMap[self._ColumnNameList[-1]]
                elif (dashcount == 2) and (len(pieces) == 2):
                    concount += 1
                    self._RowMap[pieces[1]] = int(pieces[0])
                    self._RowNameList.append(pieces[1])
                    assert len(self._RowNameList)-1 == \
                        self._RowMap[self._RowNameList[-1]]

        (ObjObject,
         ConstraintMap,
         remaining_lpfile_rows,
         AllConstraintNames) = \
            self._read_parse_lp(self._lpfilename)

        os.remove(self._lpfilename)

        #print("@@@@@@@@@@@ COLS @@@@@@@@@@@@")
        #print(set(self._ColumnNameList)-set(self._AllVars))
        #print("@@@@@@@@@@@ ROWS @@@@@@@@@@@@")
        #print(set(self._RowNameList)-set(AllConstraintNames))

        # ONE_VAR_CONSTANT always appears as an extra variable in the LP file
        assert len(self._ColumnMap)-1 == len(self._AllVars)

        # c_e_ONE_VAR_CONSTANT only appears as an extra constraint when
        # the objective function has a constant term
        assert len(self._RowMap) == \
            len(AllConstraintNames) + \
            self._count_constraint_ONE_VAR_CONSTANT
        #print sorted(ObjObject.VarToCoeff.keys())
        #print [name for name, col in \
        #       sorted(self._ColumnMap.items(),key=itemgetter(1))]
        assert set(ObjObject.VarToCoeff.keys()) == \
            set([name for name, col in \
                 sorted(list(self._ColumnMap.items()),
                        key=itemgetter(1))][:len(ObjObject.VarToCoeff)])

        if global_reference_scenario:
            (MatrixEntries_ConstrToRow_Map,
             SecondStageConstrOrder,
             self._num_first_stage_constraints,
             self._num_second_stage_constraints) = \
                self.sort_write_lp(ph,
                                   ObjObject,
                                   ConstraintMap,
                                   StageToConstraintMap,
                                   remaining_lpfile_rows)
        else:
            # for the matrix.sc file we need to know which constraint is
            # in which row:
            MatrixEntries_ConstrToRow_Map = {}
            FirstStage = StageToConstraintMap['FirstStage']
            ConstrNames = set(ConstraintMap.keys())
            FirstStageConstrOrder = sorted(ConstrNames.intersection(FirstStage))
            SecondStageConstrOrder = sorted(ConstrNames - set(FirstStage))

            # so that we know in which rows the constraints with
            # stochastic data are (first row has index 0)
            self._num_first_stage_constraints = 0
            self._num_second_stage_constraints = 0
            count_rows = -1
            for name in FirstStageConstrOrder:
                count_rows += 1
                MatrixEntries_ConstrToRow_Map[name] = count_rows
                self._num_first_stage_constraints += 1
            for name in SecondStageConstrOrder:
                count_rows += 1
                MatrixEntries_ConstrToRow_Map[name] = count_rows
                self._num_second_stage_constraints += 1

        if not getattr(self._reference_scenario_instance,
                       _disable_stoch_rhs_flagname,
                       _disable_stoch_rhs_default):
            self._num_stochastic_rhs = 0
            self._write_rhs_sc(ph,
                               ConstraintMap,
                               SecondStageConstrOrder)

        if not getattr(self._reference_scenario_instance,
                       _disable_stoch_costs_flagname,
                       _disable_stoch_costs_default):
            self._num_stochastic_costs = 0
            self._write_obj_sc(ph, ObjObject)

        if not getattr(self._reference_scenario_instance,
                       _disable_stoch_matrix_flagname,
                       _disable_stoch_matrix_default):
            self._num_stochastic_matrix_entries = 0
            self._write_matrix_sc(
                ph,
                ConstraintMap,
                SecondStageConstrOrder,
                MatrixEntries_ConstrToRow_Map,
                include_position_section=global_reference_scenario)

        if global_reference_scenario:
            return (self.input_file_name_list,
                    self._FirstStageVars,
                    self._FirstStageVarIdMap,
                    self._FirstStageDerivedVars,
                    self._FirstStageDerivedVarIdMap,
                    self._SecondStageVars,
                    self._SecondStageVarIdMap,
                    self._StageCostVarStageMap,
                    self._num_stochastic_rhs,
                    self._num_stochastic_matrix_entries,
                    self._num_stochastic_costs,
                    self._num_first_stage_constraints,
                    self._num_second_stage_constraints,
                    self._count_constraint_ONE_VAR_CONSTANT)

    # Write the lp file for a scenario and return the maximum
    # character count for names in the file
    def _write_reference_scenario_lp(self, ph):

        # Make sure the pyomo plugins are loaded
        import pyomo.environ
        lp_file_writer = pyomo.repn.plugins.cpxlp.ProblemWriter_cpxlp()

        # Write the LP file
        print(("Writing LP file to %s" % (self._lpfilename,)))
        scenariotree_vars = \
            self._reference_scenario_instance.\
                _ScenarioTreeSymbolMap.bySymbol
        rootnode_vars = \
            ph._scenario_tree.findRootNode()._standard_variable_ids
        firststage_ids = \
            set(id(vardata) for scenariotree_id, vardata \
                    in iteritems(scenariotree_vars) \
                    if scenariotree_id in rootnode_vars)
        capabilities = lambda x: True
        text_labeler = TextLabeler()
        labeler = lambda x: text_labeler(x) + \
                  (""
                   if ((not isinstance(x, _VarData)) or \
                       (id(x) not in firststage_ids)) else \
                   self._firststage_var_suffix)
        output_filename, symbol_map = \
            lp_file_writer(self._reference_scenario_instance,
                           self._lpfilename,
                           capabilities,
                           {'labeler' : labeler})
        # Plugins no longer need/support calls to deactivate()
        #lp_file_writer.deactivate()
        assert output_filename == self._lpfilename
        #print("\nModel written\n")

        # Return the max symbol length (for siphelp)
        max_name_len = max(len(symbol) for symbol in symbol_map.bySymbol)
        return max_name_len, symbol_map

    def sort_write_lp(self, ph, ObjObject, ConstraintMap, StageToConstraintMap, remaining_lpfile_rows):

        try:
            #print("\nWrite the LP file for dd in sorted_LPfile.lp\n")
            lp = open("sorted_LPfile.lp", "wt")
        except IOError:
            print("IO Error so that sorted_LPfile.lp cannot be created.")
            sys.out(1)

        # keep track of input file names
        self.input_file_name_list.append('sorted_LPfile.lp')

        # for the matrix.sc file we need to know which constraint is
        # in which row:
        MatrixEntries_ConstrToRow_Map = {}

        lp.write(ObjObject[0]+"\n "+ObjObject[1]+":\n")
        self.print_coeff_var_from_map(ObjObject[2], lp)

        # assume blank line before and after constraint indicator
        lp.write("\ns.t.\n")

        FirstStage = StageToConstraintMap['FirstStage']
        ConstrNames = set(ConstraintMap.keys())
        FirstStageConstrOrder = sorted(ConstrNames.intersection(FirstStage))
        SecondStageConstrOrder = sorted(ConstrNames - set(FirstStage))

        # so that we know in which rows the constraints with
        # stochastic data are (first row has index 0)
        count_rows = -1

        self._num_first_stage_constraints = 0
        self._num_second_stage_constraints = 0
        for name in FirstStageConstrOrder:
            lp.write("\n"+name+":\n")
            count_rows += 1
            MatrixEntries_ConstrToRow_Map[name] = count_rows
            lp_con = ConstraintMap[name]
            self.print_coeff_var_from_map(lp_con[0], lp)
            lp.write(lp_con[1]+" "+lp_con[2]+"\n")
            self._num_first_stage_constraints += 1

        for name in SecondStageConstrOrder:
            lp.write("\n"+name+":\n")
            count_rows += 1
            MatrixEntries_ConstrToRow_Map[name] = count_rows
            lp_con = ConstraintMap[name]
            self.print_coeff_var_from_map(lp_con[0], lp)
            lp.write(lp_con[1]+" "+lp_con[2]+"\n")
            self._num_second_stage_constraints += 1

        # print the remaining rows of the lp file
        for row in range(len(remaining_lpfile_rows)):
            lp.write("\n")
            for i in range(len(remaining_lpfile_rows[row])):
                lp.write(remaining_lpfile_rows[row][i]+" ")

        lp.close()

        return MatrixEntries_ConstrToRow_Map, SecondStageConstrOrder, self._num_first_stage_constraints, self._num_second_stage_constraints

    def _read_parse_lp(self, lp_filename):

        # read and parse the lp file which was generated by
        # lp_file_writer above in post_ph_initialization(self, ph)

        # This class has the OBJ sense, the OBJ name and a
        # variable-value-dictionary
        ObjObject = LPFileObjClass()

        # The keys are the constraint names, value is a class with a
        # variable-value-dictionary, the comparator and the RHS.
        ConstraintMap = {}

        try:
            #print(("\nName of the lp file which will be read: "+str(lp_filename)+"\n"))
#dlw Dec2014            lp_file = csv.reader(open(lp_filename, 'rb'), delimiter=' ', quotechar='|', skipinitialspace=True)
            lp_file = csv.reader(open(lp_filename, 'rt'), delimiter=' ', quotechar='|', skipinitialspace=True)
        except csv.Error as e:
            print((lp_filename+" is not found!"))
            sys.exit('file %s, line %d: %s' % (filename, f.line_num, e))

        # increment this only after you have processed the row
        list_lp_file = list(lp_file)[0:]

        CurrRow = 0
        CurrRow = self.SkipLPNullLines(CurrRow, list_lp_file)

        # assume it is a sense
        ObjObject.AssignSense(list_lp_file[CurrRow][0])
        CurrRow += 1

        CurrElem = 0
        # now assume we are on to the name
        name = self.ClearName(list_lp_file[CurrRow][CurrElem])
        ObjObject.AssignName(name)
        CurrRow, CurrElem = \
            self.LocalPostionUpdate(list_lp_file, CurrRow, CurrElem, 1)

        # send the pairs to the obj object
        # assume the obj is followed by a blank line!!!
        while list_lp_file[CurrRow] != []:
            LastCoeff = None
            # process the pairs
            if LastCoeff is not None:
                ObjObject.AddToMap(LastCoeff, list_lp_file[CurrRow][CurrElem])
                LastCoeff = None
                CurrRow, CurrElem = self.LocalPostionUpdate(list_lp_file, CurrRow, CurrElem, 1)
            elif 1+CurrElem == len(list_lp_file[CurrRow]):
                LastCoeff = list_lp_file[CurrRow][CurrElem]
                CurrRow, CurrElem = self.LocalPostionUpdate(list_lp_file, CurrRow, CurrElem, 1)
            else:
                ObjObject.AddToMap(list_lp_file[CurrRow][CurrElem], list_lp_file[CurrRow][CurrElem+1])
                LastCoeff = None
                CurrRow, CurrElem = self.LocalPostionUpdate(list_lp_file, CurrRow, CurrElem, 2)

        if LastCoeff is not None:
            print("\nerror mismatched pairs in objective function")
            exit(1)

        #print("ObjObject = "+str(ObjObject()))

        CurrRow = self.SkipLPNullLines(CurrRow, list_lp_file)
        CurrElem = 0
        # check if we reached the constraints
        if (('s.t.' in list_lp_file[CurrRow][CurrElem]) or \
            ('st' in list_lp_file[CurrRow][CurrElem]) or \
            ('subject to' in list_lp_file[CurrRow][CurrElem])):

            CurrRow += 1
            CurrRow = self.SkipLPNullLines(CurrRow, list_lp_file)

            while list_lp_file[CurrRow][0].strip() != 'bounds':

                # If this constraint appears in the LP file it means the
                # objective included constant terms
                if "c_e_ONE_VAR_CONSTANT" in list_lp_file[CurrRow][0]:
                    self._count_constraint_ONE_VAR_CONSTANT = 1
                    break

                CurrConstraintName = self.ClearName(list_lp_file[CurrRow][CurrElem])
                ConstraintMap[CurrConstraintName] = LPFileConstraintClass()
                if len(list_lp_file[CurrRow]) > 1:
                    CurrElem += 1
                else:
                    CurrRow += 1

                LastCoeff = None
                end = ''
                while list_lp_file[CurrRow] != []:
                    # process the pairs, keep track of last coefficient and
                    # variable, just in case we reach the end of the
                    # constraint (AssignComparator and AssignRHS need them)
                    if LastCoeff is not None:
                        end = ConstraintMap[CurrConstraintName].AddToMap(LastCoeff, list_lp_file[CurrRow][CurrElem])
                        LastVar = list_lp_file[CurrRow][CurrElem]
                        #print "in 1. IF: LastCoeff, LastVar =", LastCoeff, LastVar; sys.stdout.flush()
                        if end != 'EndOfConstraint':
                            LastCoeff = None
                            CurrRow, CurrElem = self.LocalPostionUpdate(list_lp_file, CurrRow, CurrElem, 1)
                    elif 1+CurrElem == len(list_lp_file[CurrRow]):
                        LastCoeff = list_lp_file[CurrRow][CurrElem]
                        CurrRow, CurrElem = self.LocalPostionUpdate(list_lp_file, CurrRow, CurrElem, 1)
                        #print "in 1. ElIF: LastCoeff=", LastCoeff; sys.stdout.flush()
                    elif 2+CurrElem <= len(list_lp_file[CurrRow]):
                        end = ConstraintMap[CurrConstraintName].AddToMap(list_lp_file[CurrRow][CurrElem], list_lp_file[CurrRow][CurrElem+1])
                        LastCoeff = list_lp_file[CurrRow][CurrElem]
                        LastVar = list_lp_file[CurrRow][CurrElem+1]
                        #print "in 2. ElIF: LastCoeff, LastVar =", LastCoeff, LastVar; sys.stdout.flush()
                        if end != 'EndOfConstraint':
                            LastCoeff = None
                            CurrRow, CurrElem = self.LocalPostionUpdate(list_lp_file, CurrRow, CurrElem, 2)
                    if end == 'EndOfConstraint':
                        #print "\nend of constraint"
                        ConstraintMap[CurrConstraintName].AssignComparator(LastCoeff)
                        ConstraintMap[CurrConstraintName].AssignRHS(LastVar)
                        #print "in 2. IF: LastCoeff, LastVar =", LastCoeff, LastVar; sys.stdout.flush()
                        CurrRow += 2
                        CurrElem = 0
                        break

        if self._count_constraint_ONE_VAR_CONSTANT:
            # if this constraint appears make sure it appears as
            # the last constraint in the LP file and will be
            # included with the remaining_lpfile_rows
            assert "c_e_ONE_VAR_CONSTANT" in list_lp_file[CurrRow][0]
            assert CurrRow < len(list_lp_file)

        remaining_lpfile_rows = []
        while CurrRow < len(list_lp_file):
            remaining_lpfile_rows.append(list_lp_file[CurrRow])
            CurrRow += 1
        AllConstraintNames = list(ConstraintMap.keys())

        return ObjObject, ConstraintMap, remaining_lpfile_rows, AllConstraintNames

    def print_coeff_var_from_map(self, VarToCoeffMap, lp_file):
        keys = list(VarToCoeffMap.keys())
        #string sorting
        keys.sort()
        line_template = "%+"+_precision_string+" %s\n"
        for key in keys:
            lp_file.write(line_template % (VarToCoeffMap[key], key))

    def SkipLPNullLines(self, CurrRow, lp_file):
        # skip over comment lines and blank lines (assuming comment
        # character leads)
        while (lp_file[CurrRow] == [] or lp_file[CurrRow][0].startswith("\\")):
            CurrRow += 1
            if CurrRow >= len(lp_file):
                print("Error: past end of file while skipping null lines")
                sys.exit(1)
        return CurrRow

    def ClearName(self, name):
        if name[-1] == ':':
            name = name[:-1]
        return name

    def LocalPostionUpdate(self, lp_file, CurrRow, CurrElem, step):
        if ((step !=1) and (step != 2)):
            print(("unexpected step length  ="+str(step)))
        CurrElem += step
        if CurrElem > len(lp_file[CurrRow])-1:
            CurrRow += 1
            CurrElem = 0
        return CurrRow, CurrElem

    def _Populate_StageVars(self, ph, LP_symbol_map):

        all_vars_cnt = 0
        piecewise_blocks = []
        for block in self._reference_scenario_instance.block_data_objects(active=True):
            all_vars_cnt += len(list(block.component_data_objects(Var, descend_into=False)))
            if isinstance(block, (Piecewise, _PiecewiseData)):
                piecewise_blocks.append(block)

        rootnode = ph._scenario_tree.findRootNode()

        stagetwo = ph._scenario_tree._stages[1]
        leafnode = self._reference_scenario._leaf_node
        LP_byObject = LP_symbol_map.byObject
        for scenario_tree_id, vardata in \
              iteritems(self._reference_scenario_instance.\
              _ScenarioTreeSymbolMap.bySymbol):
            if vardata.is_expression_type():
                continue
            try:
                LP_name = LP_byObject[id(vardata)]
            except KeyError:
                raise ValueError("Variable with name '%s' was declared "
                                 "on the scenario tree but did not appear "
                                 "in the reference scenario LP file."
                                 % (vardata.name))
            if scenario_tree_id in rootnode._standard_variable_ids:
                self._FirstStageVars.append(LP_name)
                self._FirstStageVarIdMap[LP_name] = scenario_tree_id
            elif (scenario_tree_id in rootnode._derived_variable_ids):
                self._FirstStageDerivedVars.append(LP_name)
                self._FirstStageDerivedVarIdMap[LP_name] = scenario_tree_id
            elif (scenario_tree_id in leafnode._variable_ids):
                self._SecondStageVars.append(LP_name)
                self._SecondStageVarIdMap[LP_name] = scenario_tree_id
            else:
                print(("%s %s" % (str(scenario_tree_id), str(vardata.name))))
                # More than two stages?
                assert False
            self._AllVars.append(LP_name)

        for stage in ph._scenario_tree._stages:
            cost_variable_name, cost_variable_index = \
                stage.nodes[0]._cost_variable
            stage_cost_component = \
                self._reference_scenario_instance.\
                find_component(cost_variable_name)
            if stage_cost_component.ctype is not Expression:
                LP_name = LP_byObject[id(stage_cost_component[cost_variable_index])]
                assert LP_name not in self._FirstStageVars
                if LP_name not in self._AllVars:
                    assert LP_name not in self._SecondStageVars
                    self._SecondStageVars.append(LP_name)
                    self._StageCostVarStageMap[LP_name] = stage._name
                    self._AllVars.append(LP_name)

        # The *ONLY* case where we allow variables to exist on the
        # model that were not declared on the scenario tree is when
        # they are autogenerated by a Piecewise component

        # For now we just assume all auxiliary Piecewise variables
        # are SecondStage
        for block in piecewise_blocks:
            for vardata in block.component_data_objects(Var,
                                                        active=True,
                                                        descend_into=False):
                LP_name = LP_byObject[id(vardata)]
                self._SecondStageVars.append(LP_name)
                self._SecondStageVarIdMap[LP_name] = scenario_tree_id
                self._AllVars.append(LP_name)

        # Make sure every variable on the model has been
        # declared on the scenario tree
        if len(self._AllVars) != all_vars_cnt:
            print("**** THERE IS A PROBLEM ****")
            print("Not all model variables are on the scenario tree. Investigating...")
            print("len(self._AllVars)=", len(self._AllVars), "all_vars_cnt=", all_vars_cnt)
            all_vars = set()
            tmp_buffer = {}
            for block in self._reference_scenario_instance.block_data_objects(active=True):
                all_vars.update(vardata.getname(True, tmp_buffer) \
                                for vardata in block.component_data_objects(Var, descend_into=False))
            print(("Number of Variables Found on Model: "+str(len(all_vars))))
            print ("writing all_vars.dat")
            with open("allvars.dat",'w') as f:
                f.write("allvars.dat\n")
                for av in all_vars:
                    f.write((str(av)+"\n"))
            tree_vars = set()
            for scenario_tree_id, vardata in \
                iteritems(self._reference_scenario_instance.\
                          _ScenarioTreeSymbolMap.bySymbol):
                tree_vars.add(vardata.name)
            print(("Number of Scenario Tree Variables (found in ddsip LP file): "+str(len(tree_vars))))
            print ("writing tree_vars.dat")
            with open("tree_vars.dat",'w') as f:
                f.write("tree_vars.dat\n")
                for tv in tree_vars:
                    f.write((str(tv)+"\n"))
            cost_vars = set()
            for stage in ph._scenario_tree._stages:
                cost_variable_name, cost_variable_index = \
                    stage.nodes[0]._cost_variable
                stage_cost_component = \
                    self._reference_scenario_instance.\
                    find_component(cost_variable_name)
                if stage_cost_component.ctype is not Expression:
                    cost_vars.add(stage_cost_component[cost_variable_index].name)
            print(("Number of Scenario Tree Cost Variables (found in ddsip LP file): "+str(len(cost_vars))))
            print ("writing cost_vars.dat")
            with open("cost_vars.dat","w") as f:
                f.write("cost_vars.dat\n")
                for cv in cost_vars:
                    f.write((str(cv)+"\n"))
            print("Variables Missing from Scenario Tree (or LP file):")
            MissingSet = all_vars-(tree_vars+cost_vars)
            for ims in MissingSet:
                print ("    ",ims)
            raise ValueError("Missing scenario tree variable declarations")

        # A necessary but not sufficient sanity check to make sure the
        # second stage variable sets are the same for all
        # scenarios. This is not required by pysp, but I think this
        # assumption is made in the rest of the code here
        for tree_node in stagetwo._tree_nodes:
            assert len(leafnode._variable_ids) == \
                len(tree_node._variable_ids)

        assert len(ph._scenario_tree._stages) == 2

        # we are doing string-sort
        self._FirstStageVars.sort()
        self._SecondStageVars.sort()
        self._AllVars.sort()

    def _Constraints_Stages(self, ph, LP_symbol_map):

        # save the output in a map: key is constraint name, value is a
        # constraintindex-stage-map
        #ConstraintToStageMap = {}
        # inverse map to ConstraintToStageMap: key is stage name,
        # value is constraintname (which includes the index)
        StageToConstraintMap = {}
        # auxiliary list for StageToConstraintMap
        FirstStageConstrNameToIndex = []
        # auxiliary list for StageToConstraintMap
        SecondStageConstrNameToIndex = []

        stage1 = ph._scenario_tree._stages[0]
        stage2 = ph._scenario_tree._stages[1]

        reference_instance = self._reference_scenario_instance
        LP_byObject = LP_symbol_map.byObject
        # deal with the fact that the LP writer prepends constraint
        # names with things like 'c_e_', 'c_l_', etc depending on the
        # constraint bound type and will even split a constraint into
        # two constraints if it has two bounds

        reference_scenario = self._reference_scenario

        LP_reverse_alias = dict()
        for symbol in LP_symbol_map.bySymbol:
            LP_reverse_alias[symbol] = []
        for alias, obj_weakref in iteritems(LP_symbol_map.aliases):
            LP_reverse_alias[LP_byObject[id(obj_weakref())]].append(alias)
        for block in reference_instance.block_data_objects(active=True):
            block_repn = getattr(block, "_repn", None)
            if block_repn is None:
                raise ValueError("Unable to find _repn ComponentMap "
                                 "on block %s" % (block.name))
            isPiecewise = False
            if isinstance(block, (Piecewise, _PiecewiseData)):
                isPiecewise = True
            for constraint_data in block.component_data_objects(
                    SOSConstraint,
                    active=True,
                    descend_into=False):
                raise TypeError("SOSConstraints are not handled by the "
                                "DDSIP interface: %s"
                                % (constraint_data.name))
            for constraint_data in block.component_data_objects(
                    Constraint,
                    active=True,
                    descend_into=False):
                LP_name = LP_byObject[id(constraint_data)]
                # if it is a range constraint this will account for
                # that fact and hold and alias for each bound
                LP_aliases = LP_reverse_alias[LP_name]
                assert len(LP_aliases) > 0
                if not isPiecewise:
                    constraint_node = reference_scenario.constraintNode(
                        constraint_data,
                        repn=block_repn.get(constraint_data),
                        instance=reference_instance)
                    stage_index = reference_scenario.node_stage_index(constraint_node)
                else:
                    stage_index = 1
                if stage_index == 0:
                    FirstStageConstrNameToIndex.extend(LP_aliases)
                elif stage_index == 1:
                    SecondStageConstrNameToIndex.extend(LP_aliases)
                else:
                    # More than two stages?
                    assert False

        StageToConstraintMap['FirstStage'] = FirstStageConstrNameToIndex
        StageToConstraintMap['SecondStage'] = SecondStageConstrNameToIndex

        return StageToConstraintMap

    def _write_rhs_sc(self, ph, ConstraintMap, SecondStageConstrOrder):

        with open(self._rhsfilename, 'wt') as f:

            if isinstance(ph, _PHSolverServer):
                probability = ph._uncompressed_scenario_tree.get_scenario(
                    self._reference_scenario._name)._probability
            else:
                probability = self._reference_scenario._probability
            header = ("scenario"+str(self._scenario_index)+"\n"
                      +str(probability)+"\n")
            f.write(header)

            self._num_stochastic_rhs = 0
            for name in SecondStageConstrOrder:
                f.write((("%"+_precision_string+"\n")
                         % (float(ConstraintMap[name][2]))))
                self._num_stochastic_rhs += 1

    def _write_obj_sc(self, ph, ObjObject):

        with open(self._objfilename, 'wt') as f:

            f.write("scenario"+str(self._scenario_index)+"\n")
            self._num_stochastic_costs = 0
            varnames_appearing = sorted(ObjObject.VarToCoeff.keys())
            for name in self._ColumnMap:
                if name not in self._FirstStageVars:
                    if name in varnames_appearing:
                        f.write(("%"+_precision_string+"\n")
                                % (ObjObject.VarToCoeff[name]))
                    else:
                        f.write("0\n")
                    self._num_stochastic_costs += 1

    def _write_matrix_sc(self,
                         ph,
                         ConstraintMap,
                         SecondStageConstrOrder,
                         MatrixEntries_ConstrToRow_Map,
                         include_position_section=False):

        with open(self._matfilename, 'wt') as f:

            if include_position_section:
                f.write('position\n')
                for conname in SecondStageConstrOrder:
                    # either it follows with what ddsip help writes in
                    # the rows+cols files
                    row = self._RowMap[conname]
                    # or it follows our ordering based off of
                    # stochastic rows following first-stage rows
                    ###row = MatrixEntries_ConstrToRow_Map[conname]
                    # or something else entirely ?????

                    var_coeffs = ConstraintMap[conname].VarToCoeff
                    for varname in sorted(var_coeffs):
                        f.write("%s\n%s\n"
                                % (row,
                                   self._ColumnMap[varname]))

            self._num_stochastic_matrix_entries = 0
            f.write("scenario"+str(self._scenario_index)+"\n")
            for conname in SecondStageConstrOrder:
                # either it follows with what ddsip ddhelp writes in
                # the rows+cols files
                row = self._RowMap[conname]
                # or it follows our ordering based off of
                # stochastic rows following first-stage
                ###row = MatrixEntries_ConstrToRow_Map[conname]
                # or something else entirely ?????

                var_coeffs = ConstraintMap[conname].VarToCoeff
                for varname in sorted(var_coeffs):
                    f.write(("%"+_precision_string+"\n")
                            % (var_coeffs[varname]))
                    self._num_stochastic_matrix_entries += 1
            print(("%s %s" % list(map(str,(self._reference_scenario._name, self._num_stochastic_matrix_entries)))))

class ddextension(pyomo.common.plugin.SingletonPlugin):

    pyomo.common.plugin.implements(phextension.IPHExtension)

    def __init__(self):

        self._lpfilename = "ddlpfile.lp"
        self._firststage_var_suffix = '__DDSIP_FIRSTSTAGE'
        self._reference_scenario = None

        # keep track of input file names
        self.input_file_name_list = []
        self._FirstStageVars = []
        self._FirstStageVarIdMap = {}
        self._FirstStageDerivedVars = []
        self._FirstStageDerivedVarIdMap = {}
        self._SecondStageVars = []
        self._SecondStageVarIdMap = {}
        self._StageCostVarStageMap = {}
        self._num_stochastic_rhs = None
        self._num_stochastic_matrix_entries = None
        self._num_stochastic_costs = None
        self._num_first_stage_constraints = None
        self._num_second_stage_constraints = None
        self._count_constraint_ONE_VAR_CONSTANT = 0

    def reset(self, ph):
        self.__init__()

    def pre_ph_initialization(self, ph):
        pass

    def post_instance_creation(self, ph):
        pass

    def post_ph_initialization(self, ph):

        print("Hello from the post_ph_initialization callback in ddphextension")

        self._ScenarioVector = \
            sorted(ph._scenario_tree._scenario_map.keys())

        self._reference_scenario = \
            ph._scenario_tree._scenario_map[self._ScenarioVector[0]]
        reference_scenario_name = self._reference_scenario._name

        if ph._scenario_tree.contains_bundles():
            print("** The DDSIP interface is ignoring scenario bundles **")

        print(("\nUsing %s as reference scenario" % (reference_scenario_name)))

        if isinstance(ph._solver_manager,
                      pyomo.solvers.plugins.smanager.phpyro.SolverManager_PHPyro):

            if ph._scenario_tree.contains_bundles():

                scenario_to_bundle_map = {}
                for scenario_index, scenario_name in enumerate(self._ScenarioVector, 1):
                    for bundle in ph._scenario_tree._scenario_bundles:
                        if scenario_name in bundle._scenario_names:
                            scenario_to_bundle_map[scenario_name] = bundle._name

                ah = transmit_external_function_invocation_to_worker(
                    ph,
                    scenario_to_bundle_map[reference_scenario_name],
                    thisfile,
                    "Write_DDSIP_Input",
                    function_args=(reference_scenario_name,
                                   self._ScenarioVector.index(reference_scenario_name)+1,
                                   self._firststage_var_suffix),
                    function_kwds={'global_reference_scenario':True},
                    return_action_handle=True)
                ph._solver_manager.wait_all([ah])
                (self.input_file_name_list,
                 self._FirstStageVars,
                 self._FirstStageVarIdMap,
                 self._FirstStageDerivedVars,
                 self._FirstStageDerivedVarIdMap,
                 self._SecondStageVars,
                 self._SecondStageVarIdMap,
                 self._StageCostVarStageMap,
                 self._num_stochastic_rhs,
                 self._num_stochastic_matrix_entries,
                 self._num_stochastic_costs,
                 self._num_first_stage_constraints,
                 self._num_second_stage_constraints,
                 self._count_constraint_ONE_VAR_CONSTANT) = \
                    ph._solver_manager.get_results(ah)

                assert os.path.exists(os.path.join(os.getcwd(),'rows+cols'))
                ahs = []
                for scenario_index, scenario_name in enumerate(self._ScenarioVector, 1):
                    if scenario_name != reference_scenario_name:
                        ahs.append(transmit_external_function_invocation_to_worker(
                            ph,
                            scenario_to_bundle_map[scenario_name],
                            thisfile,
                            "Write_DDSIP_Input",
                            function_args=(scenario_name,
                                           scenario_index,
                                           self._firststage_var_suffix),
                            function_kwds={'global_reference_scenario':False},
                            return_action_handle=True))
                ph._solver_manager.wait_all(ahs)

            else:

                ah = transmit_external_function_invocation_to_worker(
                    ph,
                    reference_scenario_name,
                    thisfile,
                    "Write_DDSIP_Input",
                    function_args=(reference_scenario_name,
                                   self._ScenarioVector.index(reference_scenario_name)+1,
                                   self._firststage_var_suffix),
                    function_kwds={'global_reference_scenario':True},
                    return_action_handle=True)
                ph._solver_manager.wait_all([ah])
                (self.input_file_name_list,
                 self._FirstStageVars,
                 self._FirstStageVarIdMap,
                 self._FirstStageDerivedVars,
                 self._FirstStageDerivedVarIdMap,
                 self._SecondStageVars,
                 self._SecondStageVarIdMap,
                 self._StageCostVarStageMap,
                 self._num_stochastic_rhs,
                 self._num_stochastic_matrix_entries,
                 self._num_stochastic_costs,
                 self._num_first_stage_constraints,
                 self._num_second_stage_constraints,
                 self._count_constraint_ONE_VAR_CONSTANT) = \
                    ph._solver_manager.get_results(ah)

                assert os.path.exists(os.path.join(os.getcwd(),'rows+cols'))
                ahs = []
                for scenario_index, scenario_name in enumerate(self._ScenarioVector, 1):
                    if scenario_name != reference_scenario_name:
                        ahs.append(transmit_external_function_invocation_to_worker(
                            ph,
                            scenario_name,
                            thisfile,
                            "Write_DDSIP_Input",
                            function_args=(scenario_name,
                                           scenario_index,
                                           self._firststage_var_suffix),
                            function_kwds={'global_reference_scenario':False},
                            return_action_handle=True))
                ph._solver_manager.wait_all(ahs)
        else:
            (self.input_file_name_list,
             self._FirstStageVars,
             self._FirstStageVarIdMap,
             self._FirstStageDerivedVars,
             self._FirstStageDerivedVarIdMap,
             self._SecondStageVars,
             self._SecondStageVarIdMap,
             self._StageCostVarStageMap,
             self._num_stochastic_rhs,
             self._num_stochastic_matrix_entries,
             self._num_stochastic_costs,
             self._num_first_stage_constraints,
             self._num_second_stage_constraints,
             self._count_constraint_ONE_VAR_CONSTANT) = \
                Write_DDSIP_Input(ph, ph._scenario_tree, self._reference_scenario,
                                  reference_scenario_name,
                                  self._ScenarioVector.index(reference_scenario_name)+1,
                                  self._firststage_var_suffix,
                                  global_reference_scenario=True)

            for scenario_index, scenario_name in enumerate(self._ScenarioVector, 1):
                if scenario_name != reference_scenario_name:
                    Write_DDSIP_Input(ph, ph._scenario_tree, ph._scenario_tree.get_scenario(scenario_name),
                                      scenario_name,
                                      scenario_index,
                                      self._firststage_var_suffix,
                                      global_reference_scenario=False)

        if self._num_stochastic_rhs is None:
            print("")
            print(("Detected "+_disable_stoch_rhs_flagname+"=True"))
            print("Constraint rhs entries are assumed to be deterministic.")
        else:
            assert self._num_stochastic_rhs > 0
            rhs_filename = 'rhs.sc'
            RHS_file = open(rhs_filename,'wt')
            RHS_file.close()
            assert self._ScenarioVector[0] == reference_scenario_name
            os.system('cat '+rhs_filename+"."+reference_scenario_name+" >> "+rhs_filename)
            os.remove(rhs_filename+"."+reference_scenario_name)
            for scenario_name in self._ScenarioVector[1:]:
                assert os.path.exists(rhs_filename+"."+scenario_name)
                os.system('cat '+rhs_filename+"."+scenario_name+" >> "+rhs_filename)
                os.remove(rhs_filename+"."+scenario_name)
            # keep track of input file names
            self.input_file_name_list.append(rhs_filename)

        if self._num_stochastic_costs is None:
            print("")
            print(("Detected "+_disable_stoch_costs_flagname+"=True"))
            print("Cost terms are assumed to be deterministic.")
        else:
            assert self._num_stochastic_costs > 0
            obj_filename = 'cost.sc'
            OBJ_file = open(obj_filename,'wt')
            OBJ_file.close()
            assert self._ScenarioVector[0] == reference_scenario_name
            os.system('cat '+obj_filename+"."+reference_scenario_name+" >> "+obj_filename)
            os.remove(obj_filename+"."+reference_scenario_name)
            for scenario_name in self._ScenarioVector[1:]:
                assert os.path.exists(obj_filename+"."+scenario_name)
                os.system('cat '+obj_filename+"."+scenario_name+" >> "+obj_filename)
                os.remove(obj_filename+"."+scenario_name)
            # keep track of input file names
            self.input_file_name_list.append(obj_filename)

        if self._num_stochastic_matrix_entries is None:
            print("")
            print(("Detected "+_disable_stoch_matrix_flagname+"=True"))
            print("Constraint matrix entries are assumed to be deterministic.")
        else:
            assert self._num_stochastic_matrix_entries > 0
            mat_filename = 'matrix.sc'
            MAT_file = open(mat_filename,'wt')
            MAT_file.close()
            assert self._ScenarioVector[0] == reference_scenario_name
            os.system('cat '+mat_filename+"."+reference_scenario_name+" >> "+mat_filename)
            os.remove(mat_filename+"."+reference_scenario_name)
            for scenario_name in self._ScenarioVector[1:]:
                assert os.path.exists(mat_filename+"."+scenario_name)
                os.system('cat '+mat_filename+"."+scenario_name+" >> "+mat_filename)
                os.remove(mat_filename+"."+scenario_name)
            # keep track of input file names
            self.input_file_name_list.append(mat_filename)

        self.write_sip_in(ph)
        self.write_input2sip(ph)
        print("ddextension post_ph_initialization callback done")

    def post_iteration_0_solves(self, ph):
        pass

    def post_iteration_0(self, ph):
        pass

    def pre_iteration_k_solves(self, ph):
        pass

    def post_iteration_k_solves(self, ph):
        pass

    def post_iteration_k(self, ph):
        pass

    def post_ph_execution(self, ph):

        self.write_start_in(ph)
        # Write this file again now that we have weights
        # and variable starts
        self.write_input2sip(ph)
        #sipout = 'sipout_good'
        #os.system('rm -rf '+sipout)
        #os.system(_ddsip_exe+' < 2sip')
        #assert os.path.exists(sipout) and os.path.isdir(sipout)
        #assert os.path.exists(os.path.join(sipout,'solution.out'))
        #self._load_ddsip_solution(ph, os.path.join(sipout,'solution.out'))

    def write_sip_in(self, ph):
        try:
            print("\n\nWrite dd input file: sip.in \n")
            sipin = open("sip.in", "wt")
        except IOError:
            print("sip.in cannot be created!")
            sys.exit(1)

        # keep track of input file names
        self.input_file_name_list.append('sip.in')

        NumberOfFirstStageVars = len(self._FirstStageVars)
        # Account for the extra ONE_VAR_CONSTANT variable that
        # shows up in every Pyomo LP file
        NumberOfSecondStageVars = (len(self._SecondStageVars) + \
                                   len(self._FirstStageDerivedVars) + \
                                   1)

        NumberOfStochasticRHS = self._num_stochastic_rhs
        if NumberOfStochasticRHS is None:
            NumberOfStochasticRHS = 0

        NumberOfStochasticCosts = self._num_stochastic_costs
        if NumberOfStochasticCosts is None:
            NumberOfStochasticCosts = 0

        NumberOfStochasticMatrixEntries = self._num_stochastic_matrix_entries
        if NumberOfStochasticMatrixEntries is None:
            NumberOfStochasticMatrixEntries = 0

        NumberOfFirstStageConstraints = self._num_first_stage_constraints
        # Account for the c_e_ONE_VAR_CONSTANT constraint if it appeared
        # (i.e., the objective had a constant term)
        NumberOfSecondStageConstraints = (self._num_second_stage_constraints + \
                                          self._count_constraint_ONE_VAR_CONSTANT)
        
        NumberOfScenarios = len(self._ScenarioVector)

        sipin.write('BEGIN \n\n\n')
        sipin.write('FIRSTCON '+str(NumberOfFirstStageConstraints)+'\n')
        sipin.write('FIRSTVAR '+str(NumberOfFirstStageVars)+'\n')
        sipin.write('SECCON '+str(NumberOfSecondStageConstraints)+'\n')
        sipin.write('SECVAR '+str(NumberOfSecondStageVars)+'\n')
        sipin.write('POSTFIX '+self._firststage_var_suffix+'\n')
        sipin.write('SCENAR '+str(NumberOfScenarios)+'\n')

        sipin.write('STOCRHS '+str(NumberOfStochasticRHS)+'\n')
        sipin.write('STOCCOST '+str(NumberOfStochasticCosts)+'\n')
        sipin.write('STOCMAT '+str(NumberOfStochasticMatrixEntries)+'\n')

        sipin.write("\n\nCPLEXBEGIN\n")
        sipin.write('1035 0 * Output on screen indicator\n')
        sipin.write('2008 0.001 * Absolute Gap\n')
        sipin.write('2009 0.001 * Relative Gap\n')
        sipin.write('1039 1200 * Time limit\n')
        sipin.write('1016 1e-9 * simplex feasibility tolerance\n')
        sipin.write('1014 1e-9 * simplex optimality tolerance\n')
        sipin.write('1065 40000 * Memory available for working storage\n')
        sipin.write('2010 1e-20 * integrality tolerance\n')
        sipin.write('2008 0 * Absolute gap\n')
        sipin.write('2020 0 * Priority order\n')
        sipin.write('2012 4 * MIP display level\n')
        sipin.write('2053 2 * disjunctive cuts\n')
        sipin.write('2040 2 * flow cover cuts\n')
        sipin.write('2060 3 *DiveType mip strategy dive (probe=3)\n')
        sipin.write('CPLEXEND\n\n')

        sipin.write('MAXINHERIT 15\n')
        sipin.write('OUTLEV 5 * Debugging\n')
        sipin.write('OUTFIL 2\n')
        sipin.write('STARTI 0  * (1 to use the starting values from PH)\n')
        sipin.write('NODELI 2000 * Sipdual node limit\n')
        sipin.write('TIMELIMIT 964000 * Sipdual time limit\n')
        sipin.write('HEURISTIC 99 3 7 * Heuristics: Down, Up, Near, Common, Byaverage ...(12)\n')
        sipin.write('ABSOLUTEGAP 0.001 * Absolute duality gap allowed in DD\n')
        sipin.write('EEVPROB 1\n')
        sipin.write('RELATIVEGAP 0.01 * (0.02) Relative duality gap allowed in DD\n')
        sipin.write('BRADIRECTION -1 * Branching direction in DD\n')
        sipin.write('BRASTRATEGY 1 * Branching strategy in DD (1 = unsolved nodes first, 0 = best bound)\n')
        sipin.write('EPSILON 1e-13 * Branch epsilon for cont. var.\n')
        sipin.write('ACCURACY 5e-16 * Accuracy\n')
        sipin.write('BOUSTRATEGY 1 * Bounding strategy in DD\n')
        sipin.write('NULLDISP 1e-16\n')
        sipin.write('RELAXF 0\n')
        sipin.write('INTFIRST 1 * Branch first on integer\n')
        sipin.write('HOTSTART 4 * use previous solution as integer starting info\n')

        sipin.write('\n\nRISKMO 0 * Risk Model\n')
        sipin.write('RISKALG 1\n')
        sipin.write('WEIGHT 1\n')
        sipin.write('TARGET 54 * target if needed\n')
        sipin.write('PROBLEV .8 * probability level\n')
        sipin.write('RISKBM 11000000 * big M in \n')

        sipin.write('\n\nCBFREQ 0 * (50) Conic Bundle in every ith node\n')
        sipin.write('CBITLIM 20 * (10) Descent iteration limit for conic bundle method\n')
        sipin.write('CBTOTITLIM 50 * (1000) Total iteration limit for conic bundle method\n')
        sipin.write('NONANT 1 * Non-anticipativity representation\n')
        sipin.write('DETEQU 1 * Write Deterministic Equivalent\n')

        sipin.write('\n\nEND\n')

        sipin.close()

    def write_start_in(self, ph):
        ### Note: as of Dec 26, 2014, the ph object we have here does not have the
        ###       scenario tree that has the solution, so this will not work
        ###       I put this code in a solution writer to get what I need - DLW
        # keep track of input file names -- or not (dlw Dec 2014)
        ### self.input_file_name_list.append('start.in')

        # keep track of input file names
        self.input_file_name_list.append('start.in')

        print("\n\nWrite dd input file: start.in (** integers will be rounded ** )\n")

        # assume two-stage and we only want the first stage vars
        rootnode = ph._scenario_tree.findRootNode()
        """
        # note: DDSIP seems to get the names in alpha order from cplex
        VNames = []
        VVals = {}
        for variable_id in rootnode._variable_ids:
            var_name, index = rootnode._variable_ids[variable_id]
            name = str(var_name)+str(index)
            name = name.replace('.','_') # lp files are what matters here
            name = name.replace(',',':')
            name = name.replace(' ','')
            VNames.append(name)
            ### VVals[name] = rootnode._solution[variable_id]
            VVals[name] = rootnode.get_variable_value(var_name, index)
        VNames.sort()
        print ("Writing init solution:")
        with open('start.in', 'wt') as f:
           f.write("SOLUTION\n")
           for name in VNames:
                if rootnode.is_variable_discrete(variable_id):
                    val = str(round(float(VVals[name])))
                else:
                    val = str(float(VVals[name]))
                f.write(val+"\n")
                # aside: rounding changes slightly from Python 2.x to 3.x
        """

        rootnode_name = ph._scenario_tree.findRootNode()._name
        weights_vectors = []
        for scenarioname in self._ScenarioVector:
            scenario_weights = \
                ph._scenario_tree._scenario_map[scenarioname]._w[rootnode_name]
            scenario_weights_vector = []
            for varname_LP in self._FirstStageVars:
                scenario_tree_id = self._FirstStageVarIdMap[varname_LP]
                scenario_weights_vector.append(scenario_weights[scenario_tree_id])
            weights_vectors.append(scenario_weights_vector)

        # transpose to orient data for processing across scenarios
        weights_vectors = list(zip(*weights_vectors))
        num_scenarios = float(len(self._ScenarioVector))

        with open('PHWEIGHTS.csv','wt') as f:
            f.write("varname,"+(",".join(name for name in self._ScenarioVector))+"\n")
            for var_index, vector_w in enumerate(weights_vectors):
                varname = self._FirstStageVars[var_index]
                f.write(varname+","+(",".join(repr(w) for w in vector_w))+"\n")

        with open('start.in','wt') as f:
            f.write("SOLUTION\n")
            for name in self._FirstStageVars:
                scenario_tree_id = self._FirstStageVarIdMap[name]
                if rootnode.is_variable_discrete(scenario_tree_id):
                    f.write(("%"+_precision_string+"\n")
                            % (round(rootnode._xbars[scenario_tree_id])))
                else:
                    f.write(("%"+_precision_string+"\n")
                            % (rootnode._xbars[scenario_tree_id]))

            # keep track of input file names
            ###self.input_file_name_list.append('NONANT1.in')

            # NONANT 1 case:
            v1_transpose = []
            for vector_w in weights_vectors:
                # all but the first position
                vector_w_iter = vector_w[1:]
                tmp = [-w/num_scenarios \
                       for w in vector_w_iter]
                v1_transpose.append(tmp)
            with open('NONANT1.in','wt') as fp:
                fp.write('MULTIPLIER\n')
                f.write('MULTIPLIER\n')
                # transpose and flatten
                for column in zip(*v1_transpose):
                    fp.writelines(repr(x)+'\n' for x in column)
                    f.writelines(repr(x)+'\n' for x in column)

        # NONANT 2 case:
        v2_transpose = []
        for vector_w in weights_vectors:
            w_sum = 0.0
            tmp = []
            # all but the last element
            vector_w_iter = vector_w[:-1]
            for w in vector_w_iter:
                w_sum += w
                tmp.append(w_sum/num_scenarios)
            v2_transpose.append(tmp)
        with open('NONANT2.in','wt') as f:
            f.write('MULTIPLIER\n')
            # tranpose and flatten
            for column in zip(*v2_transpose):
                f.writelines(repr(x)+'\n' for x in column)

        # NONANT 3 case:
        v3_transpose = []
        for vector_w in weights_vectors:
            wlast = vector_w[-1]
            # all but the last element
            vector_w_iter = vector_w[:-1]
            tmp = [(w-wlast)/num_scenarios for w in vector_w_iter]
            v3_transpose.append(tmp)
        with open('NONANT3.in','wt') as f:
            f.write('MULTIPLIER\n')
            # tranpose and flatten
            for column in zip(*v3_transpose):
                f.writelines(repr(x)+'\n' for x in column)

    def write_input2sip(self, ph):
        try:
            print("\n\nWrite dd input file: input2sip \n")
            f = open('2sip', 'w')
        except IOError:
            print("File input2sip cannot be created.")
            sys.exit(1)

        if 'sip.in' in self.input_file_name_list:
            f.write("sip.in\n")
        if 'sorted_LPfile.lp' in self.input_file_name_list:
            f.write("sorted_LPfile.lp\n")
        if 'model.ord' in self.input_file_name_list:
            f.write("model.ord\n")
        if 'rhs.sc' in self.input_file_name_list:
            f.write("rhs.sc\n")
        if 'cost.sc' in self.input_file_name_list:
            f.write("cost.sc\n")
        if 'matrix.sc' in self.input_file_name_list:
            f.write("matrix.sc\n")
        if 'order.dat' in self.input_file_name_list:
            f.write("order.dat\n")
        if 'start.in' in self.input_file_name_list:
            f.write("start.in\n")

        f.close()

    def _load_ddsip_solution(self, ph, filename):

        print("Loading DDSIP solution into scenario tree")

        # If stage costs are variables, then we will be able to
        # extract them from the DDSIP solution otherwise we will have
        # to evaludate the Expression component on the instances. If
        # this is parallel ph the instances will not be present so we
        # will not be able to determine individual stage costs.  In
        # either case, set them to None now just in case we can obtain
        # values for them.
        for scenario in ph._scenario_tree._scenarios:
            for stage_name in scenario._stage_costs:
                scenario._stage_costs[stage_name] = None

        root_node_name = ph._scenario_tree.findRootNode()._name
        reference_leaf_node = self._reference_scenario._leaf_node
        with open(filename) as f:
            line = f.readline()
            while line:
                # Get first stage solution
                if "1. Best Solution" in line:
                    first_stage_solution = {}
                    assert "Variable name                Value" in f.readline()
                    assert "---" in f.readline()
                    for i in range(len(self._FirstStageVarIdMap)):
                        LP_name, sol = f.readline().strip().split()
                        first_stage_solution[self._FirstStageVarIdMap[LP_name]] = float(sol)
                    for scenario in ph._scenario_tree._scenarios:
                        scenario._x[root_node_name].update(first_stage_solution)
                    f.readline()
                elif "2. Bounds" in line:
                    assert "Scenario    Lower Bound (root)    Upper Bound" in f.readline()
                    assert "---" in f.readline()
                    for i in range(len(ph._scenario_tree._scenarios)):
                        scenario_id, obj_lb, obj_ub = f.readline().strip().split()
                        scenario = ph._scenario_tree.get_scenario(self._ScenarioVector[int(scenario_id)-1])
                        scenario._objective = float(obj_ub)
                        scenario._cost = float(obj_ub)
                    f.readline()
                elif "3. Quantiles" in line:
                    # We don't use any information from this section
                    while f.readline().strip():
                        pass
                elif "4. Second-stage solutions" in line:
                    for i in range(len(ph._scenario_tree._scenarios)):
                        scenario_id = int(f.readline().strip().split()[1][:-1])
                        scenario = ph._scenario_tree.get_scenario(self._ScenarioVector[scenario_id-1])
                        scenario_firststage_solution = scenario._x[root_node_name]
                        secondstage_solution = []
                        # Add 1 for ONE_VAR_CONSTANT
                        for j in range(len(self._SecondStageVars)+1):
                            LP_name, sol = f.readline().strip().split()
                            if LP_name in self._SecondStageVarIdMap:
                                secondstage_solution.append((self._SecondStageVarIdMap[LP_name],float(sol)))
                            elif LP_name in self._FirstStageDerivedVarIdMap:
                                scenario_firststage_solution[self._FirstStageDerivedVarIdMap[LP_name]] = float(sol)
                            elif LP_name in self._StageCostVarStageMap:
                                scenario._stage_costs[self._StageCostVarStageMap[LP_name]._name] = float(sol)
                            else:
                                assert LP_name == "ONE_VAR_CONSTANT"
                        leaf_node = scenario._leaf_node
                        print(("%s %s" % list(map(str,(leaf_node._name, reference_leaf_node._name)))))
                        scenario_secondstage_solution = scenario._x[leaf_node._name]
                        for reference_node_variable_id, sol in secondstage_solution:
                            this_node_variable_id = leaf_node._name_index_to_id[reference_leaf_node._variable_ids[reference_node_variable_id]]
                            scenario_secondstage_solution[this_node_variable_id] = sol

                line = f.readline()

        print("Updating variable statistics after loading DDSIP solution")
        ph.update_variable_statistics()

        # If the scenario instances are present, push the
        # solutions from the scenario tree. Also try to
        # recover the individual stage costs if these
        # were not variables in the model (and DDSIP solution)
        for scenario in ph._scenario_tree._scenarios:
            if scenario._instance is not None:
                print(("%s %s" % list(map(str(scenario._name, scenario._instance.name)))))
                scenario.push_solution_to_instance()
                scenario.update_solution_from_instance()
                """
                for tree_node in scenario._node_list:
                    stage_name = tree_node._stage._name
                    if scenario._stage_costs[stage_name] is None:
                        cost_variable_name, cost_variable_index = \
                            tree_node._stage._cost_variable
                        stage_cost_component = scenario._instance.find_component(cost_variable_name)
                        scenario._stage_costs[stage_name] = \
                        stage_cost_component[cost_variable_index](exception=False)
                """
        warn = False
        for scenario in ph._scenario_tree._scenarios:
            print(("%s %s" % list(map(str,(scenario._name, scenario._cost)))))
            for stage_name in scenario._stage_costs:
                if scenario._stage_costs[stage_name] is None:
                    warn = True
                    break
        if warn:
            print("***WARNING*** Individual stage costs could not be "
                  "recovered from DDSIP solution. Try adding an artificial "
                  "variable to the model for each stage cost and assign this "
                  "to the stage cost in the scenario tree specification file.")

#
# Helper Classes
#

class MatrixEntriesClass(object):
    __slots__ = ('VarUniqueValueMap',)

    def __init__(self):
        # key is Variable name, value is unique number (keeps track of
        # stochastic data)
        self.VarUniqueValueMap = {}

    def __call__(self):
        return self.VarUniqueValueMap

    def __getitem__(self, index):
        if index == 0:
            return self.VarUniqueValueMap

    def AddToMap(self, file_line):
        var = file_line[2]
        uniquenumber = file_line[3]
        self.VarUniqueValueMap[var] = uniquenumber
        #print "in MatrixEntriesClass: self.VarUniqueValueMap=", self.VarUniqueValueMap

class LPFileObjClass(object):
    __slots__ = ('Sense','Name','VarToCoeff')

    def __init__(self):
        self.Sense = None
        self.Name = None
        self.VarToCoeff = {}

    def __call__(self):
        return self.Sense, self.Name, self.VarToCoeff

    def __getitem__(self, index):
        if index == 0:
            return self.Sense
        if index == 1:
            return self.Name
        if index == 2:
            return self.VarToCoeff

    # store the objective information from a LP file
    def AssignSense(self, SenseIn):
        if ((SenseIn[0] == 'min') or (SenseIn[0] == 'max')):
            print(("The obj sense in the lp file does not seem to be valid: "+str(SenseIn)))
            sys.exit(1)
        self.Sense = SenseIn

    def AssignName(self, arg):
        self.Name = arg

    def AddToMap(self, Coeff, Var):
        # check to see if Coeff is a number
        self.VarToCoeff[Var] = float(Coeff)
        #print "in OBJClass: self.VarToCoeff=", self.VarToCoeff

class LPFileConstraintClass(object):

    __slots__ = ('VarToCoeff','Comparator','RHS')

    def __init__(self):
        self.VarToCoeff = {}
        self.Comparator = ''
        self.RHS = None

    def __call__(self):
        return self.VarToCoeff, self.Comparator, self.RHS

    def __getitem__(self, index):
        if index == 0:
            return self.VarToCoeff
        if index == 1:
            return self.Comparator
        if index == 2:
            return self.RHS

    def AddToMap(self, Coeff, Var):
        # Check if Coeff is a comparator. Then we know if we reached
        # the end of the constraint.
        if Coeff in ('=', '<', '>', '<=', '>='):
            comparator = Coeff
            rhs = Var
            return "EndOfConstraint"
        else:
            self.VarToCoeff[Var] = float(Coeff)

    def AssignComparator(self, arg):
        self.Comparator = arg
        #print "self.Comparator=", self.Comparator

    def AssignRHS(self, arg):
        self.RHS = arg
        #print "self.RHS =",self.RHS



from pyomo.core import *
from pyomo.pysp.annotations import (StochasticConstraintBoundsAnnotation,
                                    StochasticConstraintBodyAnnotation,
                                    StochasticObjectiveAnnotation)

from piecewise_model import create_instance

def pysp_instance_creation_callback(scenario_name, node_names):

    model = create_instance(scenario_name)

    #
    # SMPS Related Annotations
    #
    model.stoch_rhs = StochasticConstraintBoundsAnnotation()
    for con in model.p_second_stage[1].component_data_objects(
            Constraint,
            active=True):
        model.stoch_rhs.declare(con)
    model.stoch_rhs.declare(model.r_second_stage)
    model.stoch_matrix = StochasticConstraintBodyAnnotation()
    # exercise more of the code by testing this with an
    # indexed block and a single block
    model.stoch_matrix.declare(model.c_second_stage, variables=[model.r])
    model.stoch_matrix.declare(model.p_second_stage[1])

    model.stoch_objective = StochasticObjectiveAnnotation()
    model.stoch_objective.declare(model.o)

    return model



"""Transformation to detect variables fixed by bounds and fix them."""
import textwrap
from math import fabs

from six import iteritems

from pyomo.core.base.plugin import TransformationFactory
from pyomo.common.collections import ComponentMap
from pyomo.common.config import (ConfigBlock, ConfigValue, NonNegativeFloat,
                                 add_docstring_list)
from pyomo.core.base.var import Var
from pyomo.core.expr.numvalue import value
from pyomo.core.plugins.transform.hierarchy import IsomorphicTransformation


@TransformationFactory.register(
        'contrib.detect_fixed_vars',
        doc="Detect variables that are de-facto fixed but not considered fixed.")
class FixedVarDetector(IsomorphicTransformation):
    """Detects variables that are de-facto fixed but not considered fixed.

    For each variable :math:`v` found on the model, check to see if its lower
    bound :math:`v^{LB}` is within some tolerance of its upper bound
    :math:`v^{UB}`. If so, fix the variable to the value of :math:`v^{LB}`.

    Keyword arguments below are specified for the ``apply_to`` and
    ``create_using`` functions.

    """

    CONFIG = ConfigBlock("FixedVarDetector")
    CONFIG.declare("tmp", ConfigValue(
        default=False, domain=bool,
        description="True to store the set of transformed variables and "
        "their old values so that they can be restored."
    ))
    CONFIG.declare("tolerance", ConfigValue(
        default=1E-13, domain=NonNegativeFloat,
        description="tolerance on bound equality (LB == UB)"
    ))

    __doc__ = add_docstring_list(__doc__, CONFIG)

    def _apply_to(self, instance, **kwargs):
        config = self.CONFIG(kwargs)

        if config.tmp:
            instance._xfrm_detect_fixed_vars_old_values = ComponentMap()

        for var in instance.component_data_objects(
                ctype=Var, descend_into=True):
            if var.fixed or var.lb is None or var.ub is None:
                # if the variable is already fixed, or if it is missing a
                # bound, we skip it.
                continue
            if fabs(value(var.lb) - value(var.ub)) <= config.tolerance:
                if config.tmp:
                    instance._xfrm_detect_fixed_vars_old_values[var] = \
                        var.value
                var.fix(var.lb)

    def revert(self, instance):
        """Revert variables fixed by the transformation."""
        for var, var_value in iteritems(
                instance._xfrm_detect_fixed_vars_old_values):
            var.unfix()
            var.set_value(var_value)

        del instance._xfrm_detect_fixed_vars_old_values



import pyutilib.th as unittest
import pyomo.environ as pe
from pyomo.core.expr.calculus.diff_with_pyomo import reverse_ad, reverse_sd
from pyomo.common.getGSL import find_GSL


tol = 6


def approx_deriv(expr, wrt, delta=0.001):
    numerator = 0
    wrt.value += 2*delta
    numerator -= pe.value(expr)
    wrt.value -= delta
    numerator += 8*pe.value(expr)
    wrt.value -= 2*delta
    numerator -= 8*pe.value(expr)
    wrt.value -= delta
    numerator += pe.value(expr)
    wrt.value += 2*delta
    return numerator / (12*delta)


class TestDerivs(unittest.TestCase):
    def test_prod(self):
        m = pe.ConcreteModel()
        m.x = pe.Var(initialize=2.0)
        m.y = pe.Var(initialize=3.0)
        e = m.x * m.y
        derivs = reverse_ad(e)
        symbolic = reverse_sd(e)
        self.assertAlmostEqual(derivs[m.x], pe.value(symbolic[m.x]), tol+3)
        self.assertAlmostEqual(derivs[m.y], pe.value(symbolic[m.y]), tol+3)
        self.assertAlmostEqual(derivs[m.x], approx_deriv(e, m.x), tol)
        self.assertAlmostEqual(derivs[m.y], approx_deriv(e, m.y), tol)

    def test_sum(self):
        m = pe.ConcreteModel()
        m.x = pe.Var(initialize=2.0)
        m.y = pe.Var(initialize=3.0)
        e = 2.0*m.x + 3.0*m.y - m.x*m.y
        derivs = reverse_ad(e)
        symbolic = reverse_sd(e)
        self.assertAlmostEqual(derivs[m.x], pe.value(symbolic[m.x]), tol+3)
        self.assertAlmostEqual(derivs[m.y], pe.value(symbolic[m.y]), tol+3)
        self.assertAlmostEqual(derivs[m.x], approx_deriv(e, m.x), tol)
        self.assertAlmostEqual(derivs[m.y], approx_deriv(e, m.y), tol)

    def test_div(self):
        m = pe.ConcreteModel()
        m.x = pe.Var(initialize=2.0)
        m.y = pe.Var(initialize=3.0)
        e = m.x / m.y
        derivs = reverse_ad(e)
        symbolic = reverse_sd(e)
        self.assertAlmostEqual(derivs[m.x], pe.value(symbolic[m.x]), tol+3)
        self.assertAlmostEqual(derivs[m.y], pe.value(symbolic[m.y]), tol+3)
        self.assertAlmostEqual(derivs[m.x], approx_deriv(e, m.x), tol)
        self.assertAlmostEqual(derivs[m.y], approx_deriv(e, m.y), tol)

    def test_pow(self):
        m = pe.ConcreteModel()
        m.x = pe.Var(initialize=2.0)
        m.y = pe.Var(initialize=3.0)
        e = m.x ** m.y
        derivs = reverse_ad(e)
        symbolic = reverse_sd(e)
        self.assertAlmostEqual(derivs[m.x], pe.value(symbolic[m.x]), tol+3)
        self.assertAlmostEqual(derivs[m.y], pe.value(symbolic[m.y]), tol+3)
        self.assertAlmostEqual(derivs[m.x], approx_deriv(e, m.x), tol)
        self.assertAlmostEqual(derivs[m.y], approx_deriv(e, m.y), tol)

    def test_sqrt(self):
        m = pe.ConcreteModel()
        m.x = pe.Var(initialize=2.0)
        m.y = pe.Var(initialize=3.0)
        e = pe.sqrt(m.x)
        derivs = reverse_ad(e)
        symbolic = reverse_sd(e)
        self.assertAlmostEqual(derivs[m.x], pe.value(symbolic[m.x]), tol+3)
        self.assertAlmostEqual(derivs[m.x], approx_deriv(e, m.x), tol)

    def test_exp(self):
        m = pe.ConcreteModel()
        m.x = pe.Var(initialize=2.0)
        e = pe.exp(m.x)
        derivs = reverse_ad(e)
        symbolic = reverse_sd(e)
        self.assertAlmostEqual(derivs[m.x], pe.value(symbolic[m.x]), tol+3)
        self.assertAlmostEqual(derivs[m.x], approx_deriv(e, m.x), tol)

    def test_log(self):
        m = pe.ConcreteModel()
        m.x = pe.Var(initialize=2.0)
        e = pe.log(m.x)
        derivs = reverse_ad(e)
        symbolic = reverse_sd(e)
        self.assertAlmostEqual(derivs[m.x], pe.value(symbolic[m.x]), tol+3)
        self.assertAlmostEqual(derivs[m.x], approx_deriv(e, m.x), tol)

    def test_log10(self):
        m = pe.ConcreteModel()
        m.x = pe.Var(initialize=2.0)
        e = pe.log10(m.x)
        derivs = reverse_ad(e)
        symbolic = reverse_sd(e)
        self.assertAlmostEqual(derivs[m.x], pe.value(symbolic[m.x]), tol+3)
        self.assertAlmostEqual(derivs[m.x], approx_deriv(e, m.x), tol)

    def test_sin(self):
        m = pe.ConcreteModel()
        m.x = pe.Var(initialize=2.0)
        e = pe.sin(m.x)
        derivs = reverse_ad(e)
        symbolic = reverse_sd(e)
        self.assertAlmostEqual(derivs[m.x], pe.value(symbolic[m.x]), tol+3)
        self.assertAlmostEqual(derivs[m.x], approx_deriv(e, m.x), tol)

    def test_cos(self):
        m = pe.ConcreteModel()
        m.x = pe.Var(initialize=2.0)
        e = pe.cos(m.x)
        derivs = reverse_ad(e)
        symbolic = reverse_sd(e)
        self.assertAlmostEqual(derivs[m.x], pe.value(symbolic[m.x]), tol+3)
        self.assertAlmostEqual(derivs[m.x], approx_deriv(e, m.x), tol)

    def test_tan(self):
        m = pe.ConcreteModel()
        m.x = pe.Var(initialize=2.0)
        e = pe.tan(m.x)
        derivs = reverse_ad(e)
        symbolic = reverse_sd(e)
        self.assertAlmostEqual(derivs[m.x], pe.value(symbolic[m.x]), tol+3)
        self.assertAlmostEqual(derivs[m.x], approx_deriv(e, m.x), tol)

    def test_asin(self):
        m = pe.ConcreteModel()
        m.x = pe.Var(initialize=0.5)
        e = pe.asin(m.x)
        derivs = reverse_ad(e)
        symbolic = reverse_sd(e)
        self.assertAlmostEqual(derivs[m.x], pe.value(symbolic[m.x]), tol+3)
        self.assertAlmostEqual(derivs[m.x], approx_deriv(e, m.x), tol)

    def test_acos(self):
        m = pe.ConcreteModel()
        m.x = pe.Var(initialize=0.5)
        e = pe.acos(m.x)
        derivs = reverse_ad(e)
        symbolic = reverse_sd(e)
        self.assertAlmostEqual(derivs[m.x], pe.value(symbolic[m.x]), tol+3)
        self.assertAlmostEqual(derivs[m.x], approx_deriv(e, m.x), tol)

    def test_atan(self):
        m = pe.ConcreteModel()
        m.x = pe.Var(initialize=2.0)
        e = pe.atan(m.x)
        derivs = reverse_ad(e)
        symbolic = reverse_sd(e)
        self.assertAlmostEqual(derivs[m.x], pe.value(symbolic[m.x]), tol+3)
        self.assertAlmostEqual(derivs[m.x], approx_deriv(e, m.x), tol)

    def test_nested(self):
        m = pe.ConcreteModel()
        m.x = pe.Var(initialize=2)
        m.y = pe.Var(initialize=3)
        m.p = pe.Param(initialize=0.5, mutable=True)
        e = pe.exp(m.x**m.p + 3.2*m.y - 12)
        derivs = reverse_ad(e)
        symbolic = reverse_sd(e)
        self.assertAlmostEqual(derivs[m.x], pe.value(symbolic[m.x]), tol+3)
        self.assertAlmostEqual(derivs[m.y], pe.value(symbolic[m.y]), tol+3)
        self.assertAlmostEqual(derivs[m.p], pe.value(symbolic[m.p]), tol+3)
        self.assertAlmostEqual(derivs[m.x], approx_deriv(e, m.x), tol)
        self.assertAlmostEqual(derivs[m.y], approx_deriv(e, m.y), tol)
        self.assertAlmostEqual(derivs[m.p], approx_deriv(e, m.p), tol)

    def test_expressiondata(self):
        m = pe.ConcreteModel()
        m.x = pe.Var(initialize=3)
        m.e = pe.Expression(expr=m.x * 2)

        @m.Expression([1, 2])
        def e2(m, i):
            if i == 1:
                return m.x + 4
            else:
                return m.x ** 2
        m.o = pe.Objective(expr=m.e + 1 + m.e2[1] + m.e2[2])
        derivs = reverse_ad(m.o.expr)
        symbolic = reverse_sd(m.o.expr)
        self.assertAlmostEqual(derivs[m.x], pe.value(symbolic[m.x]), tol)

    def test_multiple_named_expressions(self):
        m = pe.ConcreteModel()
        m.x = pe.Var()
        m.y = pe.Var()
        m.x.value = 1
        m.y.value = 1
        m.E = pe.Expression(expr=m.x*m.y)
        e = m.E - m.E
        derivs = reverse_ad(e)
        self.assertAlmostEqual(derivs[m.x], 0)
        self.assertAlmostEqual(derivs[m.y], 0)
        symbolic = reverse_sd(e)
        self.assertAlmostEqual(pe.value(symbolic[m.x]), 0)
        self.assertAlmostEqual(pe.value(symbolic[m.y]), 0)

    def test_external(self):
        DLL = find_GSL()
        if not DLL:
            self.skipTest('Could not find the amplgsl.dll library')

        m = pe.ConcreteModel()
        m.hypot = pe.ExternalFunction(library=DLL, function='gsl_hypot')
        m.x = pe.Var(initialize=0.5)
        m.y = pe.Var(initialize=1.5)
        e = 2 * m.hypot(m.x, m.x*m.y)
        derivs = reverse_ad(e)
        self.assertAlmostEqual(derivs[m.x], approx_deriv(e, m.x), tol)
        self.assertAlmostEqual(derivs[m.y], approx_deriv(e, m.y), tol)



"""Transformation to aggregate equal variables."""

from __future__ import division

import textwrap

from pyomo.common.collections import ComponentMap, ComponentSet
from pyomo.core.base import Block, Constraint, VarList, Objective, TransformationFactory
from pyomo.core.expr.current import ExpressionReplacementVisitor
from pyomo.core.expr.numvalue import value
from pyomo.core.plugins.transform.hierarchy import IsomorphicTransformation
from pyomo.repn import generate_standard_repn
import logging

logger = logging.getLogger('pyomo.contrib.preprocessing')


def _get_equality_linked_variables(constraint):
    """Return the two variables linked by an equality constraint x == y.

    If the constraint does not match this form, skip it.

    """
    if value(constraint.lower) != 0 or value(constraint.upper) != 0:
        # LB and UB on constraint must be zero; otherwise, return empty tuple.
        return ()
    if constraint.body.polynomial_degree() != 1:
        # must be a linear constraint; otherwise, return empty tuple.
        return ()

    # Generate the standard linear representation
    repn = generate_standard_repn(constraint.body)
    nonzero_coef_vars = tuple(v for i, v in enumerate(repn.linear_vars)
                              # if coefficient on variable is nonzero
                              if repn.linear_coefs[i] != 0)
    if len(nonzero_coef_vars) != 2:
        # Expect two variables with nonzero cofficient in constraint;
        # otherwise, return empty tuple.
        return ()
    if sorted(coef for coef in repn.linear_coefs if coef != 0) != [-1, 1]:
        # Expect a constraint of form x == y --> 0 == -1 * x + 1 * y;
        # otherwise, return empty tuple.
        return ()
    # Above checks are satisifed. Return the variables.
    return nonzero_coef_vars


def _fix_equality_fixed_variables(model, scaling_tolerance=1E-10):
    """Detects variables fixed by a constraint: ax=b.

    Fixes the variable to the constant value (b/a) and deactivates the relevant
    constraint.

    This sub-transformation is different than contrib.detect_fixed_vars because
    it looks for x = const rather than x.lb = x.ub.

    """
    for constraint in model.component_data_objects(
        ctype=Constraint, active=True, descend_into=True
    ):
        if not (constraint.has_lb() and constraint.has_ub()):
            # Constraint is not an equality. Skip.
            continue
        if value(constraint.lower) != value(constraint.upper):
            # Constraint is not an equality. Skip.
            continue
        if constraint.body.polynomial_degree() != 1:
            # Constraint is not linear. Skip.
            continue

        # Generate the standard linear representation
        repn = generate_standard_repn(constraint.body)
        # Generator of tuples with the coefficient and variable object for
        # nonzero coefficients.
        nonzero_coef_vars = (
            (repn.linear_coefs[i], v)
            for i, v in enumerate(repn.linear_vars)
            # if coefficient on variable is nonzero
            if repn.linear_coefs[i] != 0)
        # get the coefficient and variable object
        coef, var = next(nonzero_coef_vars)
        if next(nonzero_coef_vars, None) is not None:
            # Expect one variable with nonzero cofficient in constraint;
            # otherwise, skip.
            continue
        # Constant term on the constraint body
        const = repn.constant if repn.constant is not None else 0

        if abs(coef) <= scaling_tolerance:
            logger.warn(
                "Skipping fixed variable processing for constraint %s: "
                "%s * %s + %s = %s because coefficient %s is below "
                "tolerance of %s. Check your problem scaling." %
                (constraint.name, coef, var.name, const,
                 value(constraint.lower), coef, scaling_tolerance))
            continue

        # Constraint has form lower <= coef * var + const <= upper. We know that
        # lower = upper, so coef * var + const = lower.
        var_value = (value(constraint.lower) - const) / coef

        var.fix(var_value)
        constraint.deactivate()


def _build_equality_set(model):
    """Construct an equality set map.

    Maps all variables to the set of variables that are linked to them by
    equality. Mapping takes place using id(). That is, if you have x = y, then
    you would have id(x) -> ComponentSet([x, y]) and id(y) -> ComponentSet([x,
    y]) in the mapping.

    """
    # Map of variables to their equality set (ComponentSet)
    eq_var_map = ComponentMap()

    # Loop through all the active constraints in the model
    for constraint in model.component_data_objects(
            ctype=Constraint, active=True, descend_into=True):
        eq_linked_vars = _get_equality_linked_variables(constraint)
        if not eq_linked_vars:
            continue  # if we get an empty tuple, skip to next constraint.
        v1, v2 = eq_linked_vars
        set1 = eq_var_map.get(v1, ComponentSet((v1, v2)))
        set2 = eq_var_map.get(v2, (v2,))

        # if set1 and set2 are equivalent, skip to next constraint.
        if set1 is set2:
            continue

        # add all elements of set2 to set 1
        set1.update(set2)
        # Update all elements to point to set 1
        for v in set1:
            eq_var_map[v] = set1

    return eq_var_map


# TODO: these two functions were copied from contrib.gdp_bounds.compute_bounds
# at some point, these all should be moved into pyomo.common.
def min_if_not_None(iterable):
    """Returns the minimum among non-None elements.

    Returns None when all elements are None.

    """
    non_nones = [a for a in iterable if a is not None]
    return min(non_nones or [None])  # min( [] or [None] ) -> None


def max_if_not_None(iterable):
    """Returns the maximum among non-None elements.

    Returns None when all elements are None.

    """
    non_nones = [a for a in iterable if a is not None]
    return max(non_nones or [None])  # min( [] or [None] ) -> None


@TransformationFactory.register('contrib.aggregate_vars',
          doc="Aggregate model variables that are linked by equality constraints.")
class VariableAggregator(IsomorphicTransformation):
    """Aggregate model variables that are linked by equality constraints.

    Before:

    .. math::

        x &= y \\\\
        a &= 2x + 6y + 7 \\\\
        b &= 5y + 6 \\\\

    After:

    .. math::

        z &= x = y \\\\
        a &= 8z + 7 \\\\
        b &= 5z + 6

    .. warning:: TODO: unclear what happens to "capital-E" Expressions at this point in time.

    """

    def _apply_to(self, model, detect_fixed_vars=True):
        """Apply the transformation to the given model."""
        # Generate the equality sets
        eq_var_map = _build_equality_set(model)

        # Detect and process fixed variables.
        if detect_fixed_vars:
            _fix_equality_fixed_variables(model)

        # Generate aggregation infrastructure
        model._var_aggregator_info = Block(
            doc="Holds information for the variable aggregation "
            "transformation system.")
        z = model._var_aggregator_info.z = VarList(doc="Aggregated variables.")
        # Map of the aggregate var to the equalty set (ComponentSet)
        z_to_vars = model._var_aggregator_info.z_to_vars = ComponentMap()
        # Map of variables to their corresponding aggregate var
        var_to_z = model._var_aggregator_info.var_to_z = ComponentMap()
        processed_vars = ComponentSet()

        # TODO This iteritems is sorted by the variable name of the key in
        # order to preserve determinism. Unfortunately, var.name() is an
        # expensive operation right now.
        for var, eq_set in sorted(eq_var_map.items(),
                                  key=lambda tup: tup[0].name):
            if var in processed_vars:
                continue  # Skip already-process variables

            # This would be weird. The variable hasn't been processed, but is
            # in the map. Raise an exception.
            assert var_to_z.get(var, None) is None

            z_agg = z.add()
            z_to_vars[z_agg] = eq_set
            var_to_z.update(ComponentMap((v, z_agg) for v in eq_set))

            # Set the bounds of the aggregate variable based on the bounds of
            # the variables in its equality set.
            z_agg.setlb(max_if_not_None(v.lb for v in eq_set if v.has_lb()))
            z_agg.setub(min_if_not_None(v.ub for v in eq_set if v.has_ub()))

            # Set the fixed status of the aggregate var
            fixed_vars = [v for v in eq_set if v.fixed]
            if fixed_vars:
                # Check to make sure all the fixed values are the same.
                if any(var.value != fixed_vars[0].value
                       for var in fixed_vars[1:]):
                    raise ValueError(
                        "Aggregate variable for equality set is fixed to "
                        "multiple different values: %s" % (fixed_vars,))
                z_agg.fix(fixed_vars[0].value)

                # Check that the fixed value lies within bounds.
                if z_agg.has_lb() and z_agg.value < value(z_agg.lb):
                    raise ValueError(
                        "Aggregate variable for equality set is fixed to "
                        "a value less than its lower bound: %s < LB %s" %
                        (z_agg.value, value(z_agg.lb))
                    )
                if z_agg.has_ub() and z_agg.value > value(z_agg.ub):
                    raise ValueError(
                        "Aggregate variable for equality set is fixed to "
                        "a value greater than its upper bound: %s > UB %s" %
                        (z_agg.value, value(z_agg.ub))
                    )
            else:
                # Set the value to be the average of the values within the
                # bounds only if the value is not already fixed.
                values_within_bounds = [
                    v.value for v in eq_set if (
                        v.value is not None and
                        ((z_agg.has_lb() and v.value >= value(z_agg.lb))
                         or not z_agg.has_lb()) and
                        ((z_agg.has_ub() and v.value <= value(z_agg.ub))
                         or not z_agg.has_ub())
                    )]
                num_vals = len(values_within_bounds)
                z_agg.value = (
                    sum(val for val in values_within_bounds) / num_vals) \
                    if num_vals > 0 else None

            processed_vars.update(eq_set)

        # Do the substitution
        substitution_map = {id(var): z_var
                            for var, z_var in var_to_z.items()}
        for constr in model.component_data_objects(
            ctype=Constraint, active=True
        ):
            new_body = ExpressionReplacementVisitor(
                substitute=substitution_map
            ).dfs_postorder_stack(constr.body)
            constr.set_value((constr.lower, new_body, constr.upper))

        for objective in model.component_data_objects(
            ctype=Objective, active=True
        ):
            new_expr = ExpressionReplacementVisitor(
                substitute=substitution_map
            ).dfs_postorder_stack(objective.expr)
            objective.set_value(new_expr)

    def update_variables(self, model):
        """Update the values of the variables that were replaced by aggregates.

        TODO: reduced costs

        """
        datablock = model._var_aggregator_info
        for agg_var in datablock.z.itervalues():
            if not agg_var.stale:
                for var in datablock.z_to_vars[agg_var]:
                    var.value = agg_var.value
                    var.stale = False



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________
"""
This module defines the classes that provide an NLP interface based on
the Ampl Solver Library (ASL) implementation
"""

import os
import numpy as np
import six

from scipy.sparse import coo_matrix

import pyutilib
import pyomo
import pyomo.environ as aml
from pyomo.common.collections import ComponentMap
from pyomo.common.env import CtypesEnviron
from pyomo.contrib.pynumero.interfaces.ampl_nlp import AslNLP


__all__ = ['PyomoNLP']

# TODO: There are todos in the code below
class PyomoNLP(AslNLP):
    def __init__(self, pyomo_model):
        """
        Pyomo nonlinear program interface

        Parameters
        ----------
        pyomo_model: pyomo.environ.ConcreteModel
            Pyomo concrete model
        """
        pyutilib.services.TempfileManager.push()
        try:
            # get the temp file names for the nl file
            nl_file = pyutilib.services.TempfileManager.create_tempfile(suffix='pynumero.nl')

            # The current AmplInterface code only supports a single objective function
            # Therefore, we throw an error if there is not one (and only one) active
            # objective function. This is better than adding a dummy objective that the
            # user does not know about (since we do nnot have a good place to remove
            # this objective later
            # TODO: extend the AmplInterface and the AslNLP to correctly handle this
            # This currently addresses issue #1217
            objectives = list(pyomo_model.component_data_objects(ctype=aml.Objective, active=True, descend_into=True))
            if len(objectives) != 1:
                raise NotImplementedError('The ASL interface and PyomoNLP in PyNumero currently only support single objective'
                                          ' problems. Deactivate any extra objectives you may have, or add a dummy objective'
                                          ' (f(x)=0) if you have a square problem.')
            self._objective = objectives[0]

            # write the nl file for the Pyomo model and get the symbolMap
            fname, symbolMap = pyomo.opt.WriterFactory('nl')(pyomo_model, nl_file, lambda x:True, {})

            # create component maps from vardata to idx and condata to idx
            self._vardata_to_idx = vdidx = ComponentMap()
            self._condata_to_idx = cdidx = ComponentMap()

            # TODO: Are these names totally consistent?
            for name, obj in six.iteritems(symbolMap.bySymbol):
                if name[0] == 'v':
                    vdidx[obj()] = int(name[1:])
                elif name[0] == 'c':
                    cdidx[obj()] = int(name[1:])

            # The NL writer advertises the external function libraries
            # through the PYOMO_AMPLFUNC environment variable; merge it
            # with any preexisting AMPLFUNC definitions
            amplfunc = "\n".join(
                val for val in (
                    os.environ.get('AMPLFUNC', ''),
                    os.environ.get('PYOMO_AMPLFUNC', ''),
                ) if val)
            with CtypesEnviron(AMPLFUNC=amplfunc):
                super(PyomoNLP, self).__init__(nl_file)

            # keep pyomo model in cache
            self._pyomo_model = pyomo_model

        finally:
            # delete the nl file
            pyutilib.services.TempfileManager.pop()

    def pyomo_model(self):
        """
        Return optimization model
        """
        return self._pyomo_model

    def get_pyomo_objective(self):
        """
        Return an instance of the active objective function on the Pyomo model.
        (there can be only one)
        """
        return self._objective

    def get_pyomo_variables(self):
        """
        Return an ordered list of the Pyomo VarData objects in
        the order corresponding to the primals
        """
        # ToDo: is there a more efficient way to do this
        idx_to_vardata = {i:v for v,i in six.iteritems(self._vardata_to_idx)}
        return [idx_to_vardata[i] for i in range(len(idx_to_vardata))]

    def get_pyomo_constraints(self):
        """
        Return an ordered list of the Pyomo ConData objects in
        the order corresponding to the primals
        """
        # ToDo: is there a more efficient way to do this
        idx_to_condata = {i:v for v,i in six.iteritems(self._condata_to_idx)}
        return [idx_to_condata[i] for i in range(len(idx_to_condata))]

    def variable_names(self):
        """
        Return an ordered list of the Pyomo variable
        names in the order corresponding to the primals
        """
        pyomo_variables = self.get_pyomo_variables()
        return [v.getname() for v in pyomo_variables]

    def constraint_names(self):
        """
        Return an ordered list of the Pyomo constraint
        names in the order corresponding to internal constraint order
        """
        pyomo_constraints = self.get_pyomo_constraints()
        return [v.getname() for v in pyomo_constraints]

    def get_primal_indices(self, pyomo_variables):
        """
        Return the list of indices for the primals
        corresponding to the list of Pyomo variables provided

        Parameters
        ----------
        pyomo_variables : list of Pyomo Var or VarData objects
        """
        assert isinstance(pyomo_variables, list)
        var_indices = []
        for v in pyomo_variables:
            if v.is_indexed():
                for vd in v.values():
                    var_id = self._vardata_to_idx[vd]
                    var_indices.append(var_id)
            else:
                var_id = self._vardata_to_idx[v]
                var_indices.append(var_id)
        return var_indices

    def get_constraint_indices(self, pyomo_constraints):
        """
        Return the list of indices for the constraints
        corresponding to the list of Pyomo constraints provided

        Parameters
        ----------
        pyomo_constraints : list of Pyomo Constraint or ConstraintData objects
        """
        assert isinstance(pyomo_constraints, list)
        con_indices = []
        for c in pyomo_constraints:
            if c.is_indexed():
                for cd in c.values():
                    con_id = self._condata_to_idx[cd]
                    con_indices.append(con_id)
            else:
                con_id = self._condata_to_idx[c]
                con_indices.append(con_id)
        return con_indices

    # overloaded from NLP
    def get_obj_scaling(self):
        obj = self.get_pyomo_objective()
        scaling_suffix = self._pyomo_model.component('scaling_factor')
        if scaling_suffix and scaling_suffix.ctype is aml.Suffix and \
           obj in scaling_suffix:
            return scaling_suffix[obj]
        return None

    # overloaded from NLP
    def get_primals_scaling(self):
        scaling_suffix = self._pyomo_model.component('scaling_factor')
        if scaling_suffix and scaling_suffix.ctype is aml.Suffix:
            primals_scaling = np.ones(self.n_primals())
            for i,v in enumerate(self.get_pyomo_variables()):
                if v in scaling_suffix:
                    primals_scaling[i] = scaling_suffix[v]
            return primals_scaling
        return None

    # overloaded from NLP
    def get_constraints_scaling(self):
        scaling_suffix = self._pyomo_model.component('scaling_factor')
        if scaling_suffix and scaling_suffix.ctype is aml.Suffix:
            constraints_scaling = np.ones(self.n_constraints())
            for i,c in enumerate(self.get_pyomo_constraints()):
                if c in scaling_suffix:
                    constraints_scaling[i] = scaling_suffix[c]
            return constraints_scaling
        return None

    def extract_subvector_grad_objective(self, pyomo_variables):
        """Compute the gradient of the objective and return the entries
        corresponding to the given Pyomo variables

        Parameters
        ----------
        pyomo_variables : list of Pyomo Var or VarData objects
        """
        grad_obj = self.evaluate_grad_objective()
        return grad_obj[self.get_primal_indices(pyomo_variables)]

    def extract_subvector_constraints(self, pyomo_constraints):
        """
        Return the values of the constraints
        corresponding to the list of Pyomo constraints provided

        Parameters
        ----------
        pyomo_constraints : list of Pyomo Constraint or ConstraintData objects
        """
        residuals = self.evaluate_constraints()
        return residuals[self.get_constraint_indices(pyomo_constraints)]

    def extract_submatrix_jacobian(self, pyomo_variables, pyomo_constraints):
        """
        Return the submatrix of the jacobian that corresponds to the list
        of Pyomo variables and list of Pyomo constraints provided

        Parameters
        ----------
        pyomo_variables : list of Pyomo Var or VarData objects
        pyomo_constraints : list of Pyomo Constraint or ConstraintData objects
        """
        jac = self.evaluate_jacobian()
        primal_indices = self.get_primal_indices(pyomo_variables)
        constraint_indices = self.get_constraint_indices(pyomo_constraints)
        row_mask = np.isin(jac.row, constraint_indices)
        col_mask = np.isin(jac.col, primal_indices)
        submatrix_mask = row_mask & col_mask
        submatrix_irows = np.compress(submatrix_mask, jac.row)
        submatrix_jcols = np.compress(submatrix_mask, jac.col)
        submatrix_data = np.compress(submatrix_mask, jac.data)

        # ToDo: this is expensive - have to think about how to do this with numpy
        row_submatrix_map = {j:i for i,j in enumerate(constraint_indices)}
        for i, v in enumerate(submatrix_irows):
            submatrix_irows[i] = row_submatrix_map[v]

        col_submatrix_map = {j:i for i,j in enumerate(primal_indices)}
        for i, v in enumerate(submatrix_jcols):
            submatrix_jcols[i] = col_submatrix_map[v]

        return coo_matrix((submatrix_data, (submatrix_irows, submatrix_jcols)), shape=(len(constraint_indices), len(primal_indices)))

    def extract_submatrix_hessian_lag(self, pyomo_variables_rows, pyomo_variables_cols):
        """
        Return the submatrix of the hessian of the lagrangian that
        corresponds to the list of Pyomo variables provided

        Parameters
        ----------
        pyomo_variables_rows : list of Pyomo Var or VarData objects
            List of Pyomo Var or VarData objects corresponding to the desired rows
        pyomo_variables_cols : list of Pyomo Var or VarData objects
            List of Pyomo Var or VarData objects corresponding to the desired columns
        """
        hess_lag = self.evaluate_hessian_lag()
        primal_indices_rows = self.get_primal_indices(pyomo_variables_rows)
        primal_indices_cols = self.get_primal_indices(pyomo_variables_cols)
        row_mask = np.isin(hess_lag.row, primal_indices_rows)
        col_mask = np.isin(hess_lag.col, primal_indices_cols)
        submatrix_mask = row_mask & col_mask
        submatrix_irows = np.compress(submatrix_mask, hess_lag.row)
        submatrix_jcols = np.compress(submatrix_mask, hess_lag.col)
        submatrix_data = np.compress(submatrix_mask, hess_lag.data)

        # ToDo: this is expensive - have to think about how to do this with numpy
        submatrix_map = {j:i for i,j in enumerate(primal_indices_rows)}
        for i, v in enumerate(submatrix_irows):
            submatrix_irows[i] = submatrix_map[v]

        submatrix_map = {j:i for i,j in enumerate(primal_indices_cols)}
        for i, v in enumerate(submatrix_jcols):
            submatrix_jcols[i] = submatrix_map[v]

        return coo_matrix((submatrix_data, (submatrix_irows, submatrix_jcols)), shape=(len(primal_indices_rows), len(primal_indices_cols)))
